import numbers
import os
import unittest

import numpy as np

from rubin_scheduler.data import get_data_dir
from rubin_scheduler.utils import (
    HalfSpace,
    _angular_separation,
    angular_separation,
    arcsec_from_radians,
    basic_trixels,
    cartesian_from_spherical,
    find_htmid,
    get_all_trixels,
    half_space_from_points,
    half_space_from_ra_dec,
    intersect_half_spaces,
    level_from_htmid,
    rot_about_x,
    rot_about_y,
    rot_about_z,
    spherical_from_cartesian,
    trixel_from_htmid,
)
from rubin_scheduler.utils.htm_module import _find_htmid_fast, _find_htmid_slow


def trixel_intersects_half_space(trix, hspace):
    """
    This is a brute force method to determine whether a trixel
    is inside, or at least intersects, a halfspace.
    """
    if hspace.phi > 0.25 * np.pi:
        raise RuntimeError("trixel_intersects_half_space is not safe for " "large HalfSpaces")

    # if any of the trixel's corners are within the
    # HalfSpace, return True
    ra_rad, dec_rad = spherical_from_cartesian(hspace.vector)
    for corner in trix.corners:
        ra_rad1, dec_rad1 = spherical_from_cartesian(corner)
        if _angular_separation(ra_rad, dec_rad, ra_rad1, dec_rad1) < hspace.phi:
            return True

    # if the trixel contains the HalfSpace's center,
    # return True
    if trix.contains_pt(hspace.vector):
        return True

    sinphi = np.abs(np.sin(hspace.phi))

    # Iterate over each pair of corners (c1, c2).  For each pair,
    # construct a coordinate basis in which +z is in the
    # direction of c3, and +x is along the
    # unit vector defining c_i such that the angle
    # phi of c_j in the x,y plane is positive.  This coordinate
    # system is such that the trixel edge defined by c1, c2 is
    # now along the equator of the unit sphere.  Find the point
    # of closest approach of the HalfSpace's center to the equator.
    # If that point is between c1 and c2, return True.
    for i_c_1 in range(3):
        c1 = trix.corners[i_c_1]
        for i_c_2 in range(3):
            if i_c_2 <= i_c_1:
                continue
            c2 = trix.corners[i_c_2]
            i_c_3 = 3 - (i_c_1 + i_c_2)
            c3 = trix.corners[i_c_3]
            assert i_c_3 != i_c_2
            assert i_c_3 != i_c_1
            assert i_c_1 != i_c_2

            z_axis = np.array(
                [
                    c1[1] * c2[2] - c1[2] * c2[1],
                    c2[0] * c1[2] - c1[0] * c2[2],
                    c1[0] * c2[1] - c2[0] * c1[1],
                ]
            )
            z_axis = z_axis / np.sqrt((z_axis**2).sum())

            if np.dot(z_axis, c3) < 0.0:
                z_axis *= -1.0

            assert np.abs(1.0 - np.dot(z_axis, z_axis)) < 1.0e-10
            assert np.abs(1.0 - np.dot(c1, c1)) < 1.0e-10
            assert np.abs(1.0 - np.dot(c2, c2)) < 1.0e-10
            assert np.abs(np.dot(z_axis, c1)) < 1.0e-10
            assert np.abs(np.dot(z_axis, c2)) < 1.0e-10

            # if the dot product of the center of the HalfSpace
            # with the z axis of the new coordinate system is
            # greater than the sine of the radius of the
            # halfspace, then there is no way that the halfspace
            # intersects the equator of the unit sphere in this
            # coordinate system
            if np.abs(np.dot(z_axis, hspace.vector)) > sinphi:
                continue

            x_axis = c1
            y_axis = -1.0 * np.array(
                [
                    x_axis[1] * z_axis[2] - x_axis[2] * z_axis[1],
                    z_axis[0] * x_axis[2] - x_axis[0] * z_axis[2],
                    x_axis[0] * z_axis[1] - z_axis[0] * x_axis[1],
                ]
            )

            cos_a = np.dot(x_axis, c2)
            sin_a = np.dot(y_axis, c2)

            if sin_a < 0.0:
                x_axis = c2
                y_axis = -1.0 * np.array(
                    [
                        x_axis[1] * z_axis[2] - x_axis[2] * z_axis[1],
                        z_axis[0] * x_axis[2] - x_axis[0] * z_axis[2],
                        x_axis[0] * z_axis[1] - z_axis[0] * x_axis[1],
                    ]
                )

                cos_a = np.dot(x_axis, c1)
                sin_a = np.dot(y_axis, c1)

            assert cos_a >= 0.0
            assert sin_a >= 0.0
            assert np.abs(1.0 - cos_a**2 - sin_a**2) < 1.0e-10
            assert np.abs(np.dot(x_axis, z_axis)) < 1.0e-10
            assert np.abs(np.dot(x_axis, y_axis)) < 1.0e-10
            assert np.abs(np.dot(y_axis, z_axis)) < 1.0e-10

            x_center = np.dot(x_axis, hspace.vector)

            # if the x-coordinate of the HalfSpace's center is
            # negative, the HalfSpace is on the opposite side
            # of the unit sphere; ignore this pair c1, c2
            if x_center < 0.0:
                continue

            y_center = np.dot(y_axis, hspace.vector)

            # tan_a is the tangent of the angle between
            # the x_axis and the other trixel corner in
            # the x, y plane
            tan_a = sin_a / cos_a

            # tan_extreme is the tangent of the angle in
            # the x, y plane defining the point of closest
            # approach of the HalfSpace's center to the
            # equator.  If this point is between c1, c2,
            # return True.
            tan_extreme = y_center / x_center
            if tan_extreme > 0.0 and tan_extreme < tan_a:
                return True

    return False


class HalfSpaceTest(unittest.TestCase):
    long_message = True

    def test_half_space_contains_pt(self):
        hs = HalfSpace(np.array([0.0, 0.0, 1.0]), 0.1)
        nhs = HalfSpace(np.array([0.0, 0.0, -1.0]), -0.1)
        theta = np.arcsin(0.1)
        rng = np.random.RandomState(88)
        n_tests = 200
        ra_list = rng.random_sample(n_tests) * 2.0 * np.pi
        dec_list = rng.random_sample(n_tests) * (0.5 * np.pi - theta) + theta
        for (
            ra,
            dec,
        ) in zip(ra_list, dec_list):
            xyz = cartesian_from_spherical(ra, dec)
            self.assertTrue(hs.contains_pt(xyz))
            self.assertFalse(nhs.contains_pt(xyz))

        ra_list = rng.random_sample(n_tests) * 2.0 * np.pi
        dec_list = theta - rng.random_sample(n_tests) * (0.5 * np.pi + theta)
        for (
            ra,
            dec,
        ) in zip(ra_list, dec_list):
            xyz = cartesian_from_spherical(ra, dec)
            self.assertFalse(hs.contains_pt(xyz))
            self.assertTrue(nhs.contains_pt(xyz))

        hs = HalfSpace(np.array([1.0, 0.0, 0.0]), 0.2)
        nhs = HalfSpace(np.array([-1.0, 0.0, 0.0]), -0.2)
        theta = np.arcsin(0.2)
        ra_list = rng.random_sample(n_tests) * 2.0 * np.pi
        dec_list = rng.random_sample(n_tests) * (0.5 * np.pi - theta) + theta
        for ra, dec in zip(ra_list, dec_list):
            xyz_rot = cartesian_from_spherical(ra, dec)
            xyz = rot_about_y(xyz_rot, 0.5 * np.pi)
            self.assertTrue(hs.contains_pt(xyz))
            self.assertFalse(nhs.contains_pt(xyz))

        ra_list = rng.random_sample(n_tests) * 2.0 * np.pi
        dec_list = theta - rng.random_sample(n_tests) * (0.5 * np.pi + theta)
        for (
            ra,
            dec,
        ) in zip(ra_list, dec_list):
            xyz_rot = cartesian_from_spherical(ra, dec)
            xyz = rot_about_y(xyz_rot, 0.5 * np.pi)
            self.assertFalse(hs.contains_pt(xyz))
            self.assertTrue(nhs.contains_pt(xyz))

        vv = np.array([0.5 * np.sqrt(2), -0.5 * np.sqrt(2), 0.0])
        hs = HalfSpace(vv, 0.3)
        nhs = HalfSpace(-1.0 * vv, -0.3)
        theta = np.arcsin(0.3)
        ra_list = rng.random_sample(n_tests) * 2.0 * np.pi
        dec_list = rng.random_sample(n_tests) * (0.5 * np.pi - theta) + theta

        for ra, dec in zip(ra_list, dec_list):
            xyz_rot = cartesian_from_spherical(ra, dec)
            xyz_rot = rot_about_x(xyz_rot, 0.5 * np.pi)
            xyz = rot_about_z(xyz_rot, 0.25 * np.pi)
            self.assertTrue(hs.contains_pt(xyz))
            self.assertFalse(nhs.contains_pt(xyz))

        ra_list = rng.random_sample(n_tests) * 2.0 * np.pi
        dec_list = theta - rng.random_sample(n_tests) * (0.5 * np.pi + theta)
        for (
            ra,
            dec,
        ) in zip(ra_list, dec_list):
            xyz_rot = cartesian_from_spherical(ra, dec)
            xyz_rot = rot_about_x(xyz_rot, 0.5 * np.pi)
            xyz = rot_about_z(xyz_rot, 0.25 * np.pi)
            self.assertFalse(hs.contains_pt(xyz))
            self.assertTrue(nhs.contains_pt(xyz))

    def test_halfspace_contains_pt_scaled(self):
        """
        Test that HalfSpace.contains_pt returns the same answer
        for points on and off the unit sphere
        """
        vv = np.array([0.5 * np.sqrt(2), -0.5 * np.sqrt(2), 0.0])
        hs = HalfSpace(vv, 0.3)

        ct_inside = 0
        ct_outside = 0
        rng = np.random.RandomState(8812)
        random_pts = rng.random_sample((100, 3)) * 5.0
        for pt in random_pts:
            norm = np.sqrt(np.power(pt, 2).sum())
            self.assertGreater(np.abs(1.0 - norm), 0.01)
            unnormed_ans = hs.contains_pt(pt)
            normed_pt = pt / norm
            normed_ans = hs.contains_pt(normed_pt)
            self.assertEqual(unnormed_ans, normed_ans)

            if normed_ans:
                ct_inside += 1
            else:
                ct_outside += 1

        self.assertGreater(ct_inside, 0)
        self.assertGreater(ct_outside, 0)

    def test_halfspace_contains_trixel(self):
        # test half space that is on the equator wher N3 and S0 meet
        hs = HalfSpace(np.array([1.0, 1.0, 0.0]), 0.8)
        for tx in basic_trixels:
            status = hs.contains_trixel(basic_trixels[tx])
            msg = "Failed on %s" % tx
            if tx not in ("S0", "N3"):
                self.assertEqual(status, "outside", msg=msg)
            else:
                self.assertEqual(status, "partial", msg=msg)

        # test halfspace that is centered on vertex where S0, S3, N0, N3 meet
        hs = HalfSpace(np.array([1.0, 0.0, 0.0]), 0.8)
        for tx in basic_trixels:
            status = hs.contains_trixel(basic_trixels[tx])
            msg = "Failed on %s" % tx
            if tx not in ("S0", "S3", "N0", "N3"):
                self.assertEqual(status, "outside", msg=msg)
            else:
                self.assertEqual(status, "partial", msg=msg)

        # test halfspace fully contained in N3
        hs = HalfSpace(np.array([1.0, 1.0, 1.0]), 0.9)
        for tx in basic_trixels:
            status = hs.contains_trixel(basic_trixels[tx])
            msg = "Failed on %s" % tx
            if tx != "N3":
                self.assertEqual(status, "outside", msg=msg)
            else:
                self.assertEqual(status, "partial", msg=msg)

        # test halfspace that totally contains N3
        ra, dec = basic_trixels["N3"].get_center()
        hs = HalfSpace(np.array([1.0, 1.0, 1.0]), np.cos(0.31 * np.pi))
        for tx in basic_trixels:
            status = hs.contains_trixel(basic_trixels[tx])
            msg = "Failed on %s" % tx
            if tx == "N3":
                self.assertEqual(status, "full", msg=msg)
            elif tx in ("N1", "N2", "N0", "S0", "S1", "S3"):
                self.assertEqual(status, "partial", msg=msg)
            else:
                self.assertEqual(status, "outside", msg=msg)

    def test_half_space_eq(self):
        """
        Test that __eq__() works for HalfSpace
        """
        vv = np.array([1.0, 0.9, 2.4])
        hs1 = HalfSpace(vv, 0.1)
        hs2 = HalfSpace(2.0 * vv, 0.1)
        self.assertEqual(hs1, hs2)
        hs2 = HalfSpace(vv, 0.09)
        self.assertNotEqual(hs1, hs2)
        hs2 = HalfSpace(vv - 1.0e-4 * np.array([1.0, 0.0, 0.0]), 0.1)
        self.assertNotEqual(hs1, hs2)

    def test_find_all_trixels_radius(self):
        """
        Test the method that attempts to find all of the trixels
        inside a given half space by approximating the angular
        scale of the trixels and verifying that all returned
        trixels are within radius+angular scale of the center
        of the half space.
        """
        level = 5

        # approximate the linear angular scale (in degrees)
        # of a trixel grid using the fact that there are
        # 8*4**(level-1) trixels in the grid as per equation 2.5 of
        #
        # https://www.microsoft.com/en-us/research/wp-content/uploads/2005/09/tr-2005-123.pdf
        angular_scale = np.sqrt(4.0 * np.pi * (180.0 / np.pi) ** 2 / (8.0 * 4.0 ** (level - 1)))

        ra = 43.0
        dec = 22.0
        radius = 20.0
        half_space = half_space_from_ra_dec(ra, dec, radius)
        trixel_list = half_space.find_all_trixels(level)
        self.assertGreater(len(trixel_list), 2)

        # first, check that all of the returned trixels are
        # inside the HalfSpace
        good_htmid_list = []
        for i_limit, limits in enumerate(trixel_list):
            # verify that the tuples have been sorted by
            # htmid_min
            if i_limit > 0:
                self.assertGreater(limits[0], trixel_list[i_limit - 1][1])

            for htmid in range(limits[0], limits[1] + 1):
                test_trixel = trixel_from_htmid(htmid)
                ra_trix, dec_trix = test_trixel.get_center()
                good_htmid_list.append(htmid)
                self.assertNotEqual(half_space.contains_trixel(test_trixel), "outside")

                # check that the returned trixels are within
                # radius+angular_scale of the center of the HalfSpace
                self.assertLess(
                    angular_separation(ra, dec, ra_trix, dec_trix),
                    radius + angular_scale,
                )

        # next, verify that all of the possible trixels that
        # were not returned are outside the HalfSpace
        for base_htmid in range(8, 16):
            htmid_0 = base_htmid << 2 * (level - 1)
            self.assertEqual(level_from_htmid(htmid_0), level)
            for ii in range(2 ** (2 * level - 2)):
                htmid = htmid_0 + ii
                self.assertEqual(level_from_htmid(htmid), level)
                if htmid not in good_htmid_list:
                    test_trixel = trixel_from_htmid(htmid)
                    self.assertEqual(half_space.contains_trixel(test_trixel), "outside")
                    ra_trix, dec_trix = test_trixel.get_center()
                    self.assertGreater(angular_separation(ra, dec, ra_trix, dec_trix), radius)

    def test_find_all_trixels_brute(self):
        """
        Use the method trixel_intersects_half_space defined at the
        top of this script to verify that HalfSpace.find_all_trixels works
        """
        level = 7
        trixel_dict = get_all_trixels(level)
        all_htmid = []
        for htmid in trixel_dict.keys():
            if level_from_htmid(htmid) == level:
                all_htmid.append(htmid)

        hspace = half_space_from_ra_dec(36.0, 22.1, 2.0)

        # make sure that the two methods of determining if
        # a HalfSpace contains a trixel (HalfSpace.contains_trixel
        # and trixel_interects_half_space) agree
        for htmid in all_htmid:
            trix = trixel_dict[htmid]
            msg = "offending htmid %d" % htmid
            if trixel_intersects_half_space(trix, hspace):
                self.assertNotEqual(hspace.contains_trixel(trix), "outside", msg=msg)
            else:
                self.assertEqual(hspace.contains_trixel(trix), "outside", msg=msg)

        trixel_limits = hspace.find_all_trixels(level)
        intersecting_htmid = set()

        # check that all of the trixels included in the limits
        # do, in fact, intersect or exist in the HalfSpace
        for lim in trixel_limits:
            for htmid in range(lim[0], lim[1] + 1):
                trix = trixel_dict[htmid]
                self.assertTrue(trixel_intersects_half_space(trix, hspace))
                intersecting_htmid.add(htmid)

        # check that all of the trixels not included in the limits
        # are, in fact, outside of the HalfSpace
        self.assertLess(len(intersecting_htmid), len(all_htmid))
        self.assertGreater(len(intersecting_htmid), 0)
        for htmid in all_htmid:
            if htmid in intersecting_htmid:
                continue
            trix = trixel_dict[htmid]
            self.assertFalse(trixel_intersects_half_space(trix, hspace))

    def test_half_space_from_points(self):
        rng = np.random.RandomState(88)
        for ii in range(10):
            pt1 = (rng.random_sample() * 360.0, rng.random_sample() * 180.0 - 90.0)
            pt2 = (rng.random_sample() * 360.0, rng.random_sample() * 180.0 - 90.0)
            pt3 = (rng.random_sample() * 360.0, rng.random_sample() * 180.0 - 90.0)
            hs = half_space_from_points(pt1, pt2, pt3)

            # check that the HalfSpace contains pt3
            vv3 = cartesian_from_spherical(np.radians(pt3[0]), np.radians(pt3[1]))
            self.assertTrue(hs.contains_pt(vv3))

            # check that the HalfSpace encompasses 1/2 of the unit sphere
            self.assertAlmostEqual(hs.phi, 0.5 * np.pi, 10)
            self.assertAlmostEqual(hs.dd, 0.0, 10)

            # check that pt1 and pt2 are 90 degrees away from the center
            # of the HalfSpace
            vv1 = cartesian_from_spherical(np.radians(pt1[0]), np.radians(pt1[1]))
            vv2 = cartesian_from_spherical(np.radians(pt2[0]), np.radians(pt2[1]))
            self.assertAlmostEqual(np.dot(vv1, hs.vector), 0.0, 10)
            self.assertAlmostEqual(np.dot(vv2, hs.vector), 0.0, 10)

    def test__half_space_intersection(self):
        # Test that the two roots of an intersection are the
        # correct angular distance from the centers of the
        # half spaces
        ra1 = 22.0
        dec1 = 45.0
        rad1 = 10.0
        ra2 = 23.5
        dec2 = 37.9
        rad2 = 9.2
        hs1 = half_space_from_ra_dec(ra1, dec1, rad1)
        hs2 = half_space_from_ra_dec(ra2, dec2, rad2)
        roots = intersect_half_spaces(hs1, hs2)
        self.assertEqual(len(roots), 2)
        self.assertAlmostEqual(np.sqrt(np.sum(roots[0] ** 2)), 1.0, 10)
        self.assertAlmostEqual(np.sqrt(np.sum(roots[1] ** 2)), 1.0, 10)
        ra_r1, dec_r1 = np.degrees(spherical_from_cartesian(roots[0]))
        ra_r2, dec_r2 = np.degrees(spherical_from_cartesian(roots[1]))
        dd = angular_separation(ra1, dec1, ra_r1, dec_r1)
        self.assertAlmostEqual(dd, rad1, 10)
        dd = angular_separation(ra1, dec1, ra_r2, dec_r2)
        self.assertAlmostEqual(dd, rad1, 10)
        dd = angular_separation(ra2, dec2, ra_r1, dec_r1)
        self.assertAlmostEqual(dd, rad2, 10)
        dd = angular_separation(ra2, dec2, ra_r2, dec_r2)
        self.assertAlmostEqual(dd, rad2, 10)

        # test that two non-intersecting HalfSpaces return no roots
        hs1 = half_space_from_ra_dec(0.0, 90.0, 1.0)
        hs2 = half_space_from_ra_dec(20.0, -75.0, 5.0)
        roots = intersect_half_spaces(hs1, hs2)
        self.assertEqual(len(roots), 0)

        # test that two half spaces that are inside each other
        # return no roots
        hs1 = half_space_from_ra_dec(77.0, 10.0, 20.0)
        hs2 = half_space_from_ra_dec(75.0, 8.0, 0.2)
        roots = intersect_half_spaces(hs1, hs2)
        self.assertEqual(len(roots), 0)

        # test that two half spaces with identical centers
        # return no roots
        hs1 = half_space_from_ra_dec(11.0, -23.0, 1.0)
        hs2 = half_space_from_ra_dec(11.0, -23.0, 0.2)
        roots = intersect_half_spaces(hs1, hs2)
        self.assertEqual(len(roots), 0)

        roots = intersect_half_spaces(hs1, hs1)
        self.assertEqual(len(roots), 0)

    def test_merge_trixel_bounds(self):
        """
        Test that the merge_trixel_bounds method works
        """
        input_bound = [(1, 7), (2, 4), (21, 35), (8, 11), (36, 42), (43, 43)]
        result = HalfSpace.merge_trixel_bounds(input_bound)
        shld_be = [(1, 11), (21, 43)]
        self.assertEqual(result, shld_be)

    def test_join_trixel_bound_sets(self):
        """
        Test that HalfSpace.join_trixel_bound_sets works
        """
        b1 = [(32, 47), (6, 8), (11, 19), (12, 14), (66, 73)]
        b2 = [(35, 41), (7, 15), (41, 44)]
        result = HalfSpace.join_trixel_bound_sets(b1, b2)
        shld_be = [(7, 8), (11, 15), (35, 44)]
        self.assertEqual(result, shld_be)

    def test_contains_many_pts(self):
        """
        Test that HalfSpace.contains_many_pts works
        """
        rng = np.random.RandomState(5142)
        n_pts = 100
        vv_list = np.zeros((n_pts, 3), dtype=float)
        for ii in range(n_pts):
            vv = rng.random_sample(3) - 0.5
            vv /= np.sqrt(np.sum(vv**2))
            vv_list[ii] = vv

        vv = rng.random_sample(3) - 0.5
        vv /= np.sqrt(np.sum(vv**2))
        hs = HalfSpace(vv, 0.3)
        results = hs.contains_many_pts(vv_list)
        is_true = np.where(results)[0]
        self.assertGreater(len(is_true), n_pts // 4)
        self.assertLess(len(is_true), 3 * n_pts // 4)
        for i_vv, vv in enumerate(vv_list):
            self.assertEqual(hs.contains_pt(vv), results[i_vv])


class TrixelFinderTest(unittest.TestCase):
    long_message = True

    def check_pt(self, pt, answer):
        """
        Take a Cartesian point (pt) and a known
        htmid for that point (answer).  Find the htmid
        for the point using find_htmid and verify that
        we get the expected answer.
        """
        ra, dec = spherical_from_cartesian(pt)
        ii = find_htmid(np.degrees(ra), np.degrees(dec), 3)
        binary = "{0:b}".format(ii)
        self.assertEqual(binary, answer)

    def test_against_fatboy(self):
        """
        Test find_htmid against a random selection of stars from fatboy
        """
        dtype = np.dtype([("htmid", int), ("ra", float), ("dec", float)])
        data = np.genfromtxt(
            os.path.join(get_data_dir(), "utils", "htmid_test_data.txt"),
            dtype=dtype,
        )
        self.assertGreater(len(data), 20)
        for i_pt in range(len(data)):
            htmid_test = find_htmid(data["ra"][i_pt], data["dec"][i_pt], 21)
            self.assertEqual(htmid_test, data["htmid"][i_pt])
            level_test = level_from_htmid(htmid_test)
            self.assertEqual(level_test, 21)

    def test_find_htmid_vectorized(self):
        """
        Test that find_htmid works correctly on vectors
        """
        rng = np.random.RandomState(81723122)
        n_samples = 1000
        ra = rng.random_sample(n_samples) * 360.0
        dec = rng.random_sample(n_samples) * 180.0 - 90.0
        level = 7
        htmid_vec = find_htmid(ra, dec, level)
        self.assertIsInstance(htmid_vec, np.ndarray)
        htmid_fast = _find_htmid_fast(ra, dec, level)
        self.assertIsInstance(htmid_fast, np.ndarray)
        np.testing.assert_array_equal(htmid_vec, htmid_fast)
        for ii in range(n_samples):
            htmid_slow = _find_htmid_slow(ra[ii], dec[ii], level)
            self.assertIsInstance(htmid_slow, numbers.Number)
            self.assertEqual(htmid_slow, htmid_vec[ii])
            htmid_single = find_htmid(ra[ii], dec[ii], level)
            self.assertIsInstance(htmid_single, numbers.Number)
            self.assertEqual(htmid_single, htmid_vec[ii])

    def test_level_from_htmid(self):
        """
        Test that level_from_htmid behaves as expected
        """
        for ii in range(8, 16):
            self.assertEqual(level_from_htmid(ii), 1)

        self.assertEqual(level_from_htmid(2**9 + 5), 4)
        self.assertEqual(level_from_htmid(2**15 + 88), 7)

        with self.assertRaises(RuntimeError) as context:
            level_from_htmid(2**10)
        self.assertIn("4+2n", context.exception.args[0])

        for ii in range(8):
            with self.assertRaises(RuntimeError) as context:
                level_from_htmid(2**10)
            self.assertIn("4+2n", context.exception.args[0])

    def test_trixel_finding(self):
        """
        Check that find_htmid works by passing in some
        points whose htmid are known because of their
        proximity to the corners of low-level Trixels.
        Use check_pt to verify that find_htmid gives
        the right answer.
        """
        epsilon = 1.0e-6
        dx = np.array([epsilon, 0.0, 0.0])
        dy = np.array([0.0, epsilon, 0.0])
        dz = np.array([0.0, 0.0, epsilon])

        xx = np.array([1.0, 0.0, 0.0])
        yy = np.array([0.0, 1.0, 0.0])
        zz = np.array([0.0, 0.0, 1.0])

        pt = xx + dy + dz
        # N320
        self.check_pt(pt, "11111000")

        pt = xx - dy + dz
        # N000
        self.check_pt(pt, "11000000")

        pt = xx - dy - dz
        # S320
        self.check_pt(pt, "10111000")

        pt = yy + dx + dz
        # N300
        self.check_pt(pt, "11110000")

        pt = yy - dx + dz
        # N220
        self.check_pt(pt, "11101000")

        pt = yy - dx - dz
        # S100
        self.check_pt(pt, "10010000")

        pt = zz + dy + dx
        # N310
        self.check_pt(pt, "11110100")

        pt = zz - dy + dx
        # N010
        self.check_pt(pt, "11000100")

        pt = zz - dy - dx
        # N110
        self.check_pt(pt, "11010100")

        pt = -xx + dz + dy
        # N200
        self.check_pt(pt, "11100000")

        pt = -xx - dz + dy
        # S120
        self.check_pt(pt, "10011000")

        pt = -xx - dz - dy
        # S200
        self.check_pt(pt, "10100000")

        pt = -yy + dx + dz
        # N020
        self.check_pt(pt, "11001000")

        pt = -yy - dx + dz
        # N100
        self.check_pt(pt, "11010000")

        pt = -yy - dx - dz
        # S220
        self.check_pt(pt, "10101000")

        pt = -zz + dx + dy
        # S010
        self.check_pt(pt, "10000100")

        pt = -zz - dx + dy
        # S110
        self.check_pt(pt, "10010100")

        pt = -zz - dx - dy
        # S210
        self.check_pt(pt, "10100100")

        pt = xx + yy + zz
        # N333
        self.check_pt(pt, "11111111")

    def test_trixel_from_htmid(self):
        """
        Check that trixel_from_htmid works by
        finding the htmid from an RA, Dec pair,
        instantiating the Trixel corresponding
        to that htmid, and verifying that that
        Trixel (and not its neighbors) contains
        the RA, Dec pair.
        """
        rng = np.random.RandomState(88)
        n_tests = 100
        for i_test in range(n_tests):
            pt = rng.normal(0.0, 1.0, 3)
            ra, dec = spherical_from_cartesian(pt)
            ra = np.degrees(ra)
            dec = np.degrees(dec)
            ii = find_htmid(ra, dec, 5)
            tt = trixel_from_htmid(ii)
            self.assertTrue(tt.contains(ra, dec))
            tt1 = trixel_from_htmid(ii - 1)
            self.assertFalse(tt1.contains(ra, dec))
            tt2 = trixel_from_htmid(ii + 1)
            self.assertFalse(tt2.contains(ra, dec))

    def test_trixel_eq_ne(self):
        """
        Test that the __eq__ and __ne__ operators on the Trixel class work
        """
        t1 = trixel_from_htmid(8 * 16 + 1)
        t2 = trixel_from_htmid(8 * 16 + 1)
        self.assertEqual(t1, t2)
        t3 = trixel_from_htmid(8 * 16 + 3)
        self.assertNotEqual(t1, t3)
        self.assertTrue(t1 == t2)
        self.assertFalse(t1 == t3)
        self.assertTrue(t1 != t3)
        self.assertFalse(t2 == t3)
        self.assertTrue(t2 != t3)

    def test_get_all_trixels(self):
        """
        Test method to get all trixels up to a certain level
        """
        max_level = 5
        n_trixel_per_level = {}
        n_trixel_per_level[0] = 0
        for level in range(1, max_level + 1):
            n_trixel_per_level[level] = 8 * (4 ** (level - 1))

        trixel_dict = get_all_trixels(max_level)
        n_found = {}
        for level in range(max_level + 1):
            n_found[level] = 0

        for htmid in trixel_dict:
            level = level_from_htmid(htmid)
            n_found[level] += 1

        # verify that the correct number of trixels were
        # found per level
        for level in n_found:
            msg = "failed on level %d" % level
            self.assertEqual(n_found[level], n_trixel_per_level[level], msg=msg)

        # make sure no trixels were duplicated
        self.assertEqual(len(np.unique(list(trixel_dict.keys()))), len(trixel_dict))

        for htmid in trixel_dict.keys():
            level = level_from_htmid(htmid)
            self.assertLessEqual(level, max_level)
            self.assertGreaterEqual(level, 1)
            t0 = trixel_from_htmid(htmid)
            self.assertEqual(t0, trixel_dict[htmid])

    def test_trixel_bounding_circle(self):
        """
        Verify that the trixel's bounding_circle method returns
        a circle that contains all of the corners of the
        trixel
        """
        rng = np.random.RandomState(142)
        n_test_cases = 5
        for i_test in range(n_test_cases):
            htmid = (13 << 6) + rng.randint(1, 2**6 - 1)
            trixel = trixel_from_htmid(htmid)
            bounding_circle = trixel.bounding_circle
            ra_0, dec_0 = spherical_from_cartesian(bounding_circle[0])
            ra_list = []
            dec_list = []
            for cc in trixel.corners:
                ra, dec = spherical_from_cartesian(cc)
                ra_list.append(ra)
                dec_list.append(dec)
            ra_list = np.array(ra_list)
            dec_list = np.array(dec_list)
            distance = _angular_separation(ra_0, dec_0, ra_list, dec_list)
            distance = arcsec_from_radians(distance)
            radius = arcsec_from_radians(bounding_circle[2])
            self.assertLessEqual(distance.max() - radius, 1.0e-8)
            self.assertLess(np.abs(distance.max() - radius), 1.0e-8)

    def test_trixel_contains_many(self):
        """
        Test that trixel.contains_pt and trixel.contains can
        work with numpy arrays of input
        """
        htmid = (15 << 6) + 45
        trixel = trixel_from_htmid(htmid)
        ra_0, dec_0 = trixel.get_center()
        radius = trixel.get_radius()
        rng = np.random.RandomState(44)
        n_pts = 100
        rr = radius * rng.random_sample(n_pts) * 1.1
        theta = rng.random_sample(n_pts) * 2.0 * np.pi
        ra_list = ra_0 + rr * np.cos(theta)
        dec_list = dec_0 + rr * np.sin(theta)
        contains_arr = trixel.contains(ra_list, dec_list)
        n_in = 0
        n_out = 0
        for i_pt in range(n_pts):
            single_contains = trixel.contains(ra_list[i_pt], dec_list[i_pt])
            self.assertEqual(single_contains, contains_arr[i_pt])
            if single_contains:
                n_in += 1
            else:
                n_out += 1

        self.assertGreater(n_in, 0)
        self.assertGreater(n_out, 0)

        xyz_list = cartesian_from_spherical(np.radians(ra_list), np.radians(dec_list))

        contains_xyz_arr = trixel.contains_pt(xyz_list)
        np.testing.assert_array_equal(contains_xyz_arr, contains_arr)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
import pandas as pd

import rubin_scheduler.scheduler.basis_functions as basis_functions
import rubin_scheduler.scheduler.surveys as surveys
from rubin_scheduler.scheduler.model_observatory import ModelObservatory
from rubin_scheduler.scheduler.schedulers import CoreScheduler
from rubin_scheduler.scheduler.utils import generate_all_sky


class TestCoreSched(unittest.TestCase):
    def testsched(self):
        nside = 32
        # Just set up a very simple target map, dec limited, one filter
        sky_dict = generate_all_sky(nside, mask=-1)
        target_map = np.where(
            ((sky_dict["map"] >= 0) & (sky_dict["dec"] < 2) & (sky_dict["dec"] > -65)), 1, 0
        )

        bfs = []
        bfs.append(basis_functions.M5DiffBasisFunction(nside=nside))
        bfs.append(basis_functions.TargetMapBasisFunction(target_map=target_map, norm_factor=1))
        weights = np.array([1.0, 1])
        survey = surveys.GreedySurvey(bfs, weights)
        scheduler = CoreScheduler([survey])

        observatory = ModelObservatory()

        # Check that we can update conditions
        scheduler.update_conditions(observatory.return_conditions())

        # Check that we can get an observation out
        obs = scheduler.request_observation()
        assert obs is not None

        # Check that we can flush the Queue
        scheduler.flush_queue()
        assert len(scheduler.queue) == 0

        # Check that we can add an observation
        scheduler.add_observation(obs)

        # Check dunder methods
        self.assertIsInstance(repr(scheduler), str)
        self.assertIsInstance(str(scheduler), str)

        # Check access methods
        _ = scheduler.get_basis_functions([0, 0])
        _ = scheduler.get_healpix_maps([0, 0])

        # Check survey access methods
        reward_df = scheduler.make_reward_df(observatory.return_conditions())
        self.assertIsInstance(reward_df, pd.DataFrame)
        reward_df = scheduler.make_reward_df(observatory.return_conditions(), accum=False)
        self.assertIsInstance(reward_df, pd.DataFrame)

        obs = scheduler.request_observation()
        surveys_df = scheduler.surveys_df(0)
        self.assertIsInstance(surveys_df, pd.DataFrame)

        # Test we can record basis function values when requested
        recording_scheduler = CoreScheduler([survey], keep_rewards=True)
        recording_scheduler.update_conditions(observatory.return_conditions())
        obs = recording_scheduler.request_observation()
        self.assertIsInstance(recording_scheduler.queue_reward_df, pd.DataFrame)


if __name__ == "__main__":
    unittest.main()



import unittest

from astropy.time import Time, TimeDelta

from rubin_scheduler.site_models import UnscheduledDowntimeData


class UnscheduledDowntimeDataTest(unittest.TestCase):
    def setUp(self):
        self.th = Time("2020-01-01", format="isot", scale="tai")
        self.startofnight = -0.34
        self.seed = 1516231120
        self.survey_length = 3650 * 2

    def test_basic_information_after_creation(self):
        downtime_data = UnscheduledDowntimeData(
            self.th,
            start_of_night_offset=self.startofnight,
            survey_length=self.survey_length,
            seed=self.seed,
        )
        self.assertEqual(downtime_data.seed, self.seed)
        self.assertEqual(downtime_data.survey_length, self.survey_length)
        self.assertEqual(self.th + TimeDelta(self.startofnight, format="jd"), downtime_data.night0)
        downtime_data = UnscheduledDowntimeData(
            self.th,
            start_of_night_offset=0,
            survey_length=self.survey_length,
            seed=self.seed,
        )
        self.assertEqual(downtime_data.night0, self.th)

    def test_information_after_initialization(self):
        downtime_data = UnscheduledDowntimeData(
            self.th,
            start_of_night_offset=self.startofnight,
            survey_length=self.survey_length,
            seed=self.seed,
        )
        downtime_data.make_data()
        self.assertEqual(len(downtime_data.downtime), 155)
        # Check some of the downtime values.
        dnight = downtime_data.downtime["end"] - downtime_data.downtime["start"]
        self.assertEqual(dnight[0].jd, 1)
        self.assertEqual(downtime_data.downtime["activity"][0], "minor event")
        self.assertEqual(dnight[2].jd, 7)
        self.assertEqual(downtime_data.downtime["activity"][2], "major event")

    def test_alternate_seed(self):
        downtime_data = UnscheduledDowntimeData(
            self.th,
            start_of_night_offset=self.startofnight,
            survey_length=self.survey_length,
            seed=3,
        )
        downtime_data.make_data()
        self.assertEqual(len(downtime_data.downtime), 145)

    def test_call(self):
        downtime_data = UnscheduledDowntimeData(
            self.th,
            start_of_night_offset=self.startofnight,
            survey_length=self.survey_length,
            seed=self.seed,
        )
        downtime_data.make_data()
        downtimes = downtime_data()
        self.assertEqual(downtimes["activity"][2], "major event")


if __name__ == "__main__":
    unittest.main()



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

from documenteer.conf.guide import *  # noqa: F403, import *

linkcheck_retries = 2



__all__ = ("comcam_tessellate",)

import numpy as np


def comcam_tessellate(side_length=0.7, overlap=0.11):
    """Tesselate the sphere with a square footprint

    XXX--TODO:  This really sucks at the poles, should add some
    different pole cap behavior.

    Parameters
    ----------
    side_length : float (0.7)
        The length of a side of the square (degrees)
    overlap : float (0.11)
        How much overlap to have in pointings

    Returns
    -------
    fields : numpy array
       With 'RA' and 'dec' keys that have the field positions in radians
    """

    # Convert to radians for all internal work
    side_length = np.radians(side_length)
    overlap = np.radians(overlap)

    rotation_step = side_length / 2.0

    step_size = side_length - overlap
    n_dec_steps = np.ceil(np.pi / 2.0 / step_size).astype(int)
    dec_stripes = np.linspace(0, np.pi / 2.0, n_dec_steps)

    # Lists to hold the RA and dec coords
    ras = []
    decs = []
    running_step = 0.0
    test_decs = np.array([-step_size, 0.0, step_size])
    for dec in dec_stripes:
        # Find the largest circumfrance the squares will have to cover
        circum = np.max(2.0 * np.pi * np.cos(test_decs + dec))
        n_ra_steps = np.ceil(circum / step_size).astype(int)
        new_ras = np.linspace(0, 2.0 * np.pi, n_ra_steps)
        running_step += rotation_step / np.cos(dec)
        new_ras = (new_ras + running_step) % (2.0 * np.pi)
        ras.extend(new_ras.tolist())
        new_decs = np.empty(new_ras.size)
        new_decs.fill(dec)
        decs.extend(new_decs.tolist())

    # That was the northern hemisphere, copy to do the south
    north = np.where(np.array(decs) > 0)
    decs.extend((-1.0 * np.array(decs))[north].tolist())
    new_ras = -1.0 * np.array(ras)[north]
    new_ras = new_ras % (2.0 * np.pi)
    ras.extend(new_ras.tolist())

    names = ["RA", "dec"]
    types = [float, float]
    fields = np.empty(len(ras), dtype=list(zip(names, types)))
    fields["RA"] = np.array(ras)
    fields["dec"] = np.array(decs)

    return fields



__all__ = ("ddf_locations",)

# Tuples of RA,dec in degrees
ELAISS1 = (9.45, -44.0)
XMM_LSS = (35.708333, -4 - 45 / 60.0)
ECDFS = (53.125, -28.0 - 6 / 60.0)
COSMOS = (150.1, 2.0 + 10.0 / 60.0 + 55 / 3600.0)
EDFS_A = (58.90, -49.315)
EDFS_B = (63.6, -47.60)


def ddf_locations():
    """Return the DDF locations as as dict. RA and dec in degrees."""
    result = {}
    result["ELAISS1"] = ELAISS1
    result["XMM_LSS"] = XMM_LSS
    result["ECDFS"] = ECDFS
    result["COSMOS"] = COSMOS
    result["EDFS_a"] = EDFS_A
    result["EDFS_b"] = EDFS_B

    return result



__all__ = ("TooMaster", "TooSurvey")

import copy

import numpy as np

from rubin_scheduler.scheduler.surveys import BaseSurvey, BlobSurvey


class TooMaster(BaseSurvey):
    """
    A target of opportunity class. Every time a new ToO comes in, it
    will spawn a new sub-survey.

    Parameters
    ----------
    example__to_o_survey : rubin_scheduler.scheduler.surveys.ToO_survey
        A survey object that will be coppied and have a new target
        map set for each incoming ToO.
    """

    def __init__(self, example__to_o_survey):
        self.example__to_o_survey = example__to_o_survey
        self.surveys = []
        self.highest_reward = -np.inf
        self.scheduled_obs = None

    def add_observation(self, observation, indx=None):
        if len(self.surveys) > 0:
            for survey in self.surveys:
                survey.add_observation(observation, indx=indx)

    def _spawn_new_survey(self, too):
        """Create a new survey object for a ToO we haven't seen before.

        Parameters
        ----------
        too : rubin_scheduler.scheduler.utils.TargetoO object
        """
        new_survey = copy.deepcopy(self.example__to_o_survey)
        new_survey.set_id(too.id)
        new_survey.set_target_map(too.footprint)

        return new_survey

    def _check_survey_list(self, conditions):
        """There is a current ToO in the conditions."""

        running_ids = [survey.too_id for survey in self.surveys]
        current_ids = [too.id for too in conditions.targets_of_opportunity]

        # delete any ToO surveys that are no longer relevant
        self.surveys = [survey for survey in self.surveys if survey.too_id in current_ids]

        # Spawn new surveys that are needed
        new_surveys = []
        for too in conditions.targets_of_opportunity:
            if too.id not in running_ids:
                new_surveys.append(self._spawn_new_survey(too))
        self.surveys.extend(new_surveys)

    def calc_reward_function(self, conditions):
        # Catch if a new ToO has happened
        if conditions.targets_of_opportunity is not None:
            self._check_survey_list(conditions)

        if len(self.surveys) > 0:
            rewards = [np.nanmax(survey.calc_reward_function(conditions)) for survey in self.surveys]
            self.reward = np.nanmax(rewards)
            self.highest_reward = np.min(np.where(rewards == self.reward))
        else:
            self.reward = -np.inf
            self.highest_reward = None
        return self.reward

    def generate_observations(self, conditions):
        if self.reward > -np.inf:
            result = self.surveys[self.highest_reward].generate_observations(conditions)
            return result


class TooSurvey(BlobSurvey):
    """Survey class to catch incoming target of opportunity
    anouncements and try to observe them.

    The idea is that we can dynamically update the target
    footprint basis function, and add new features as more ToOs come in.

    Parameters
    ----------
    too_id : int (None)
        A unique integer ID for the ToO getting observed
    """

    def __init__(
        self,
        basis_functions,
        basis_weights,
        filtername1="r",
        filtername2=None,
        slew_approx=7.5,
        filter_change_approx=140.0,
        read_approx=2.0,
        exptime=30.0,
        nexp=2,
        ideal_pair_time=22.0,
        min_pair_time=15.0,
        search_radius=30.0,
        alt_max=85.0,
        az_range=180.0,
        flush_time=30.0,
        smoothing_kernel=None,
        nside=None,
        dither=True,
        seed=42,
        ignore_obs=None,
        survey_note="ToO",
        detailers=None,
        camera="LSST",
        too_id=None,
        survey_name=None,
    ):
        super(TooSurvey, self).__init__(
            basis_functions=basis_functions,
            basis_weights=basis_weights,
            filtername1=filtername1,
            filtername2=filtername2,
            slew_approx=slew_approx,
            filter_change_approx=filter_change_approx,
            read_approx=read_approx,
            exptime=exptime,
            nexp=nexp,
            ideal_pair_time=ideal_pair_time,
            min_pair_time=min_pair_time,
            search_radius=search_radius,
            alt_max=alt_max,
            az_range=az_range,
            flush_time=flush_time,
            smoothing_kernel=smoothing_kernel,
            nside=nside,
            dither=dither,
            seed=seed,
            ignore_obs=ignore_obs,
            survey_note=survey_note,
            detailers=detailers,
            camera=camera,
            survey_name=survey_name,
        )
        # Include the ToO id in the note
        self.survey_note_base = self.survey_note
        self.set_id(too_id)

    def set_id(self, newid):
        """Set the id"""
        self.too_id = newid
        self.survey_note = self.survey_note_base + ", " + str(newid)

    def set_target_map(self, newmap):
        """
        Expect one of the basis functions to be Footprint_nvis_basis_function
        """
        for basis_func in self.basis_functions:
            if hasattr(basis_func, "footprint"):
                basis_func.footprint = newmap

    def generate_observations_rough(self, conditions):
        # Always spin the tesselation before generating a new block.
        if self.dither:
            self._spin_fields(conditions)
        result = super(TooSurvey, self).generate_observations_rough(conditions)
        return result



"""Tools for maintaining an archive of opsim output and metadata.
"""

__all__ = [
    "make_sim_archive_dir",
    "transfer_archive_dir",
    "check_opsim_archive_resource",
    "read_archived_sim_metadata",
    "make_sim_archive_cli",
    "drive_sim",
]

import argparse
import datetime
import hashlib
import json
import logging
import lzma
import os
import pickle
import shutil
import socket
import sys
from contextlib import redirect_stdout
from numbers import Integral, Number
from pathlib import Path
from tempfile import TemporaryDirectory

import numpy as np
import pandas as pd
import yaml
from astropy.time import Time

import rubin_scheduler
from rubin_scheduler.scheduler import sim_runner
from rubin_scheduler.scheduler.utils import SchemaConverter

LOGGER = logging.getLogger(__name__)

try:
    from lsst.resources import ResourcePath
except ModuleNotFoundError:
    LOGGER.error("Module lsst.resources required to use rubin_scheduler.sim_archive.")

try:
    from conda.cli.main_list import print_packages
    from conda.gateways.disk.test import is_conda_environment

    have_conda = True
except ModuleNotFoundError:
    have_conda = False
    LOGGER.warning("No conda module found, no conda envirment data will be saved")


def make_sim_archive_dir(
    observations,
    reward_df=None,
    obs_rewards=None,
    in_files={},
    sim_runner_kwargs={},
    tags=[],
    label=None,
    data_path=None,
    capture_env=True,
):
    """Create or fill a local simulation archive directory.

    Parameters
    ----------
    observations : `numpy.recarray`
        The observations data, in the "obs" format as accepted and
        created by `rubin_scheduler.scheduler.utils.SchemaConverter`.
    reward_df : `pandas.DataFrame`, optional
        The reward data, by default None.
    obs_rewards : `pandas.DataFrame`, optional
        The observation rewards data, by default None.
    in_files : `dict`, optional
        Additional input files to be included in the archive,
        by default {}.
    sim_runner_kwargs : `dict`, optional
        Additional simulation runner keyword arguments, by default {}.
    tags : `list` [`str`], optional
        A list of tags/keywords to be included in the metadata, by
        default [].
    label : `str`, optional
        A label to be included in the metadata, by default None.
    data_path : `str` or `pathlib.Path`, optional
        The path to the simulation archive directory, by default None.
    capture_env : `bool`
        Use the current environment as the sim environment.
        Defaults to True.

    Returns
    -------
    data_dir : `pathlib.Path` or `tempfile.TemporaryDirectory`
        The temporary directory containing the simulation archive.
    """
    if data_path is None:
        data_dir = TemporaryDirectory()
        data_path = Path(data_dir.name)
    else:
        data_dir = None

        if not isinstance(data_path, Path):
            data_path = Path(data_path)

    files = {}

    # Save the observations
    files["observations"] = {"name": "opsim.db"}
    opsim_output_fname = data_path.joinpath(files["observations"]["name"])
    SchemaConverter().obs2opsim(observations, filename=opsim_output_fname)

    # Save the rewards
    if reward_df is not None and obs_rewards is not None:
        files["rewards"] = {"name": "rewards.h5"}
        rewards_fname = data_path.joinpath(files["rewards"]["name"])
        if reward_df is not None:
            reward_df.to_hdf(rewards_fname, "reward_df")
        if obs_rewards is not None:
            obs_rewards.to_hdf(rewards_fname, "obs_rewards")

    # Save basic statistics
    files["statistics"] = {"name": "obs_stats.txt"}
    stats_fname = data_path.joinpath(files["statistics"]["name"])
    with open(stats_fname, "w") as stats_io:
        print(SchemaConverter().obs2opsim(observations).describe().T.to_csv(sep="\t"), file=stats_io)

    if capture_env:
        # Save the conda environment
        conda_prefix = Path(sys.executable).parent.parent.as_posix()
        if have_conda and is_conda_environment(conda_prefix):
            conda_base_fname = "environment.txt"
            environment_fname = data_path.joinpath(conda_base_fname).as_posix()

            # Python equivilent of
            # conda list --export -p $conda_prefix > $environment_fname
            with open(environment_fname, "w") as environment_io:
                with redirect_stdout(environment_io):
                    print_packages(conda_prefix, format="export")

            files["environment"] = {"name": conda_base_fname}

        # Save pypi packages
        pypi_base_fname = "pypi.json"
        pypi_fname = data_path.joinpath(pypi_base_fname).as_posix()

        pip_json_output = os.popen("pip list --format json")
        pip_list = json.loads(pip_json_output.read())

        with open(pypi_fname, "w") as pypi_io:
            print(json.dumps(pip_list, indent=4), file=pypi_io)

        files["pypi"] = {"name": pypi_base_fname}

    # Add supplied files
    for file_type, fname in in_files.items():
        files[file_type] = {"name": Path(fname).name}
        try:
            shutil.copyfile(fname, data_path.joinpath(files[file_type]["name"]))
        except shutil.SameFileError:
            pass

    # Add file hashes
    for file_type in files:
        fname = data_path.joinpath(files[file_type]["name"])
        with open(fname, "rb") as file_io:
            content = file_io.read()

        files[file_type]["md5"] = hashlib.md5(content).hexdigest()

    def convert_mjd_to_dayobs(mjd):
        # Use dayObs defn. from SITCOMTN-32: https://sitcomtn-032.lsst.io/
        evening_local_mjd = np.floor(mjd - 0.5).astype(int)
        evening_local_iso = Time(evening_local_mjd, format="mjd").iso[:10]
        return evening_local_iso

    opsim_metadata = {}
    if capture_env:
        opsim_metadata["scheduler_version"] = rubin_scheduler.__version__
        opsim_metadata["host"] = socket.getfqdn()

    opsim_metadata["username"] = os.environ["USER"]

    simulation_dates = {}
    if "mjd_start" in sim_runner_kwargs:
        simulation_dates["first"] = convert_mjd_to_dayobs(sim_runner_kwargs["mjd_start"])

        if "survey_length" in sim_runner_kwargs:
            simulation_dates["last"] = convert_mjd_to_dayobs(
                sim_runner_kwargs["mjd_start"] + sim_runner_kwargs["survey_length"] - 1
            )
    else:
        simulation_dates["first"] = convert_mjd_to_dayobs(observations["mjd"].min())
        simulation_dates["last"] = convert_mjd_to_dayobs(observations["mjd"].max())

    if len(sim_runner_kwargs) > 0:
        opsim_metadata["sim_runner_kwargs"] = {}
        for key, value in sim_runner_kwargs.items():
            # Cast numpy number types to ints, floats, and reals to avoid
            # confusing the yaml module.
            match value:
                case bool():
                    opsim_metadata["sim_runner_kwargs"][key] = value
                case Integral():
                    opsim_metadata["sim_runner_kwargs"][key] = int(value)
                case Number():
                    opsim_metadata["sim_runner_kwargs"][key] = float(value)
                case _:
                    opsim_metadata["sim_runner_kwargs"][key] = str(value)

    opsim_metadata["simulated_dates"] = simulation_dates
    opsim_metadata["files"] = files

    if len(tags) > 0:
        for tag in tags:
            assert isinstance(tag, str), "Tags must be strings."
        opsim_metadata["tags"] = tags

    if label is not None:
        assert isinstance(label, str), "The sim label must be a string."
        opsim_metadata["label"] = label

    sim_metadata_fname = data_path.joinpath("sim_metadata.yaml")
    with open(sim_metadata_fname, "w") as sim_metadata_io:
        print(yaml.dump(opsim_metadata, indent=4), file=sim_metadata_io)

    files["metadata"] = {"name": sim_metadata_fname}

    if data_dir is not None:
        # If we created a temporary directory, if we do not return it, it
        # will get automatically cleaned up, losing our work.
        # So, return it.
        return data_dir

    return data_path


def transfer_archive_dir(archive_dir, archive_base_uri="s3://rubin-scheduler-prenight/opsim/"):
    """Transfer the contents of an archive directory to an resource.

    Parameters
    ----------
    archive_dir : `str`
        The path to the archive directory containing the files to be
        transferred.
    archive_base_uri : `str`, optional
        The base URI where the archive files will be transferred to.
        Default is "s3://rubin-scheduler-prenight/opsim/".

    Returns
    -------
    resource_rpath : `ResourcePath`
        The destination resource.
    """

    metadata_fname = Path(archive_dir).joinpath("sim_metadata.yaml")
    with open(metadata_fname, "r") as metadata_io:
        sim_metadata = yaml.safe_load(metadata_io)

    insert_date = datetime.datetime.utcnow().date().isoformat()
    insert_date_rpath = ResourcePath(archive_base_uri).join(insert_date, forceDirectory=True)
    if not insert_date_rpath.exists():
        insert_date_rpath.mkdir()

    # Number the sims in the insert date dir by
    # looing for all the interger directories, and choosing the next one.
    found_ids = []
    for base_dir, found_dirs, found_files in insert_date_rpath.walk():
        if base_dir == insert_date_rpath:
            for found_dir in found_dirs:
                try:
                    found_dir_index = found_dir[:-1] if found_dir.endswith("/") else found_dir
                    found_ids.append(int(found_dir_index))
                except ValueError:
                    pass

    new_id = max(found_ids) + 1 if len(found_ids) > 0 else 1
    resource_rpath = insert_date_rpath.join(f"{new_id}", forceDirectory=True)
    resource_rpath.mkdir()

    # Include the metadata file itself.
    sim_metadata["files"]["metadata"] = {"name": "sim_metadata.yaml"}

    for file_info in sim_metadata["files"].values():
        source_fname = Path(archive_dir).joinpath(file_info["name"])
        with open(source_fname, "rb") as source_io:
            content = source_io.read()

        destination_rpath = resource_rpath.join(file_info["name"])
        destination_rpath.write(content)

        LOGGER.info(f"Copied {source_fname} to {destination_rpath}")

    return resource_rpath


def check_opsim_archive_resource(archive_uri):
    """Check the contents of an opsim archive resource.

    Parameters
    ----------
    archive_uri : `str`
        The URI of the archive resource to be checked.

    Returns
    -------
    validity: `dict`
        A dictionary of files checked, and their validity.
    """
    metadata_path = ResourcePath(archive_uri).join("sim_metadata.yaml")
    with metadata_path.open(mode="r") as metadata_io:
        sim_metadata = yaml.safe_load(metadata_io)

    results = {}

    for file_info in sim_metadata["files"].values():
        resource_path = ResourcePath(archive_uri).join(file_info["name"])
        content = resource_path.read()

        results[file_info["name"]] = file_info["md5"] == hashlib.md5(content).hexdigest()

    return results


def _build_archived_sim_label(base_uri, metadata_resource, metadata):
    label_base = metadata_resource.dirname().geturl().removeprefix(base_uri).rstrip("/").lstrip("/")

    # If a label is supplied by the metadata, use it
    if "label" in metadata:
        label = f"{label_base} {metadata['label']}"
        return label

    try:
        sim_dates = metadata["simulated_dates"]
        first_date = sim_dates["first"]
        last_date = sim_dates["last"]
        label = f"{label_base} of {first_date}"
        if last_date != first_date:
            label = f"{label} through {last_date}"
    except KeyError:
        label = label_base

    if "scheduler_version" in metadata:
        label = f"{label} with {metadata['scheduler_version']}"

    return label


def read_archived_sim_metadata(base_uri, latest=None, num_nights=5):
    """Read metadata for a time range of archived opsim output.

    Parameters
    ----------
    base_uri : `str`
        The base URI of the archive resource to be checked.
    latest : `str`, optional
        The date of the latest simulation whose metadata should be loaded.
        This is the date on which the simulations was added to the archive,
        not necessarily the date on which the simulation was run, or any
        of the dates simulated.
        Default is today.
    num_nights : `int`
        The number of nights of the date window to load.

    Returns
    -------
    sim_metadata: `dict`
        A dictionary of metadata for simulations in the date range.
    """
    latest_mjd = int(Time.now().mjd if latest is None else Time(latest).mjd)
    earliest_mjd = int(latest_mjd - num_nights)

    all_metadata = {}
    for mjd in range(earliest_mjd, latest_mjd + 1):
        iso_date = Time(mjd, format="mjd").iso[:10]
        date_resource = ResourcePath(base_uri).join(iso_date, forceDirectory=True)
        if date_resource.exists():
            for base_dir, found_dirs, found_files in date_resource.walk(file_filter=r".*sim_metadata.yaml"):
                for found_file in found_files:
                    found_resource = ResourcePath(base_dir).join(found_file)
                    these_metadata = yaml.safe_load(found_resource.read().decode("utf-8"))
                    these_metadata["label"] = _build_archived_sim_label(
                        base_uri, found_resource, these_metadata
                    )
                    all_metadata[str(found_resource.dirname())] = these_metadata

    return all_metadata


def make_sim_archive_cli(*args):
    parser = argparse.ArgumentParser(description="Add files to sim archive")
    parser.add_argument(
        "label",
        type=str,
        help="A label for the simulation.",
    )
    parser.add_argument(
        "opsim",
        type=str,
        help="File name of opsim database.",
    )
    parser.add_argument("--rewards", type=str, default=None, help="A rewards HDF5 file.")
    parser.add_argument(
        "--scheduler_version",
        type=str,
        default=None,
        help="The version of the scheduler that producte the opsim database.",
    )
    parser.add_argument(
        "--scheduler",
        type=str,
        default=None,
        help="A snapshot of the scheduler used to produce the database, at the start of the simulation.",
    )
    parser.add_argument(
        "--script", type=str, default=None, help="The file name of the script run to create the simulation."
    )

    notebook_help = "The file name of the notebook run to create the simulation."
    notebook_help = notebook_help + " This can be produced using the %%notebook magic."
    parser.add_argument(
        "--notebook",
        type=str,
        default=None,
        help=notebook_help,
    )
    parser.add_argument(
        "--current_env",
        action="store_true",
        help="Record the current environment as the simulation environment.",
    )
    parser.add_argument(
        "--archive_base_uri",
        type=str,
        default="s3://rubin-scheduler-prenight/opsim/",
        help="Base URI for the archive",
    )
    parser.add_argument("--tags", type=str, default=[], nargs="*", help="The tags on the simulation.")
    arg_values = parser.parse_args() if len(args) == 0 else parser.parse_args(args)

    observations = SchemaConverter().opsim2obs(arg_values.opsim)

    if arg_values.rewards is not None:
        try:
            reward_df = pd.read_hdf(arg_values.rewards, "reward_df")
        except KeyError:
            reward_df = None

        try:
            obs_rewards = pd.read_hdf(arg_values.rewards, "obs_rewards")
        except KeyError:
            obs_rewards = None
    else:
        reward_df = None
        obs_rewards = None

    filename_args = ["scheduler", "script", "notebook"]
    in_files = {}
    for filename_arg in filename_args:
        try:
            filename = getattr(arg_values, filename_arg)
            if filename is not None:
                in_files[filename] = filename
        except AttributeError:
            pass

    data_path = make_sim_archive_dir(
        observations,
        reward_df,
        obs_rewards,
        in_files,
        tags=arg_values.tags,
        label=arg_values.label,
        capture_env=arg_values.current_env,
    )

    sim_archive_uri = transfer_archive_dir(data_path.name, arg_values.archive_base_uri)

    return sim_archive_uri


def drive_sim(
    observatory, scheduler, archive_uri=None, label=None, tags=[], script=None, notebook=None, **kwargs
):
    """Run a simulation and archive the results.

    Parameters
    ----------
    observatory : `ModelObservatory`
        The model for the observatory.
    scheduler : `CoreScheduler`
        The scheduler to use.
    archive_uri : `str`, optional
        The root URI of the archive resource into which the results should
        be stored. Defaults to None.
    label : `str`, optional
        The label for the simulation in the archive. Defaults to None.
    tags : `list` of `str`, optional
        The tags for the simulation in the archive. Defaults to an
        empty list.
    script : `str`
        The filename of the script producing this simulation.
        Defaults to None.
    notebook : `str`, optional
        The filename of the notebook producing the simulation.
        Defaults to None.

    Returns
    -------
    observatory : `ModelObservatory`
        The model for the observatory.
    scheduler : `CoreScheduler`
        The scheduler used.
    observations : `numpy.recarray`
        The observations produced.
    reward_df : `pandas.DataFrame`, optional
        The table of rewards. Present if `record_rewards`
        or `scheduler.keep_rewards` is True.
    obs_rewards : `pandas.Series`, optional
        The mapping of entries in reward_df to observations. Present if
        `record_rewards` or `scheduler.keep_rewards` is True.
    resource_path : `ResourcePath`, optional
        The resource path to the archive of the simulation. Present if
        `archive_uri` was set.

    Notes
    -----
    Additional parameters not described above will be passed into
    `sim_runner`.

    If the `archive_uri` parameter is not supplied, `sim_runner` is run
    directly, so that `drive_sim` can act as a drop-in replacement of
    `sim-runner`.

    In a jupyter notebook, the notebook can be saved for the notebook
    paramater using `%notebook $notebook_fname` (where `notebook_fname`
    is variable holding the filename for the notebook) in the cell prior
    to calling `drive_sim`.
    """
    if "record_rewards" in kwargs:
        if kwargs["record_rewards"] and not scheduler.keep_rewards:
            raise ValueError("To keep rewards, scheduler.keep_rewards must be True")
    else:
        kwargs["record_rewards"] = scheduler.keep_rewards

    in_files = {}
    if script is not None:
        in_files["script"] = script

    if notebook is not None:
        in_files["notebook"] = notebook

    with TemporaryDirectory() as local_data_dir:
        # We want to store the state of the scheduler at the start of
        # the sim, so we need to save it now before we run the simulation.
        scheduler_path = Path(local_data_dir).joinpath("scheduler.pickle.xz")
        with lzma.open(scheduler_path, "wb", format=lzma.FORMAT_XZ) as pio:
            pickle.dump(scheduler, pio)
            in_files["scheduler"] = scheduler_path.as_posix()

        sim_results = sim_runner(observatory, scheduler, **kwargs)

        observations = sim_results[2]
        reward_df = sim_results[3] if scheduler.keep_rewards else None
        obs_rewards = sim_results[4] if scheduler.keep_rewards else None

        data_dir = make_sim_archive_dir(
            observations,
            reward_df=reward_df,
            obs_rewards=obs_rewards,
            in_files=in_files,
            sim_runner_kwargs=kwargs,
            tags=tags,
            label=label,
            capture_env=True,
        )

        if archive_uri is not None:
            resource_path = transfer_archive_dir(data_dir.name, archive_uri)
        else:
            resource_path = ResourcePath(data_dir.name, forceDirctory=True)

    results = sim_results + (resource_path,)
    return results



__all__ = ("read_fields", "_read_fields")

import os

import numpy as np

from rubin_scheduler.data import get_data_dir
from rubin_scheduler.utils import _ra_dec_from_xyz


def _read_fields(filename=None):
    """Read in field positions.

    Parameters
    ----------
    filename : str (None)
        File to read. Defaults to icover.3.5292.23.0.txt origianlly from:
        http://neilsloane.com/icosahedral.codes/index.html

    Returns
    -------
    ra, dec : np.arrays of RA,dec values in radians
    """
    if filename is None:
        filename = os.path.join(get_data_dir(), "site_models/icover.3.5292.23.0.txt")
    cov = np.genfromtxt(filename)
    x = cov[0::3]
    y = cov[1::3]
    z = cov[2::3]

    ra, dec = _ra_dec_from_xyz(x, y, z)
    return ra, dec


def read_fields(filename=None):
    """Read in field positions.

    Parameters
    ----------
    filename : str (None)
        File to read. Defaults to icover.3.5292.23.0.txt origianlly from:
        http://neilsloane.com/icosahedral.codes/index.html

    Returns
    -------
    ra, dec : np.arrays of RA,dec values in degrees
    """
    ra, dec = _read_fields(filename=filename)
    return np.degrees(ra), np.degrees(dec)



__all__ = ("dark_sky", "dark_m5")

import os

import healpy as hp
import numpy as np

from rubin_scheduler.data import get_data_dir
from rubin_scheduler.site_models import SeeingModel
from rubin_scheduler.utils import m5_flat_sed


def dark_sky(nside=32):
    """Load an array of HEALpix maps that have the darkest expected sky
    backgrounds per filter.

    Parameters
    ----------
    nside : `int` (32)
        Desired nside resolution (default=32).

    Returns
    -------
    dark_sky_data : `np.ndarray`
        Named array with dark sky data for each band.
    """
    if not hasattr(dark_sky, "data"):
        # Load up the data
        data_dir = get_data_dir()
        data = np.load(os.path.join(data_dir, "skybrightness_pre", "dark_maps.npz"))
        dark_sky.data = data["dark_maps"].copy()
        data.close()

    dark_sky_data = np.empty(hp.nside2npix(nside), dtype=dark_sky.data.dtype)

    for band in dark_sky_data.dtype.names:
        dark_sky_data[band] = hp.pixelfunc.ud_grade(dark_sky.data[band], nside_out=nside)

    return dark_sky_data


def dark_m5(decs, filtername, latitude_rad, fiducial_FWHMEff, exptime=30.0, nexp=1):
    """Return a nominal best-depth map of the sky

    Parameters
    ----------
    decs : `float`
        The declinations for the desired points. Float or adday-like. (radians)
    filtername : `str`
        Name of filter, one of ugrizy.
    latitude_rad : `float`
        Latitude of the observatory (radians)
    fiducial_FWHMEff : `float`
        The fiducial seeing FWHMeff to use (arcsec).
    exptime : `float`
        The fiducial exposure time to assume (seconds). Default 30.
    nexp : `int`
        The number of exposures per visit. Default 1.


    """
    nside = hp.npix2nside(np.size(decs))
    ds = dark_sky(nside)[filtername]
    min_z = np.abs(decs - latitude_rad)
    airmass_min = 1 / np.cos(min_z)
    airmass_min = np.where(airmass_min < 0, np.nan, airmass_min)
    sm = SeeingModel(filter_list=[filtername])
    fwhm_eff = sm(fiducial_FWHMEff, airmass_min)["fwhmEff"][0]
    dark_m5_map = m5_flat_sed(
        filtername,
        musky=ds,
        fwhm_eff=fwhm_eff,
        exp_time=exptime,
        airmass=airmass_min,
        nexp=nexp,
        tau_cloud=0,
    )
    return dark_m5_map



import unittest

import healpy as hp
import numpy as np

import rubin_scheduler.skybrightness_pre as sbp


class TestDarkSky(unittest.TestCase):
    def test_default(self):
        dark_sky = sbp.dark_sky()

        self.assert_dark_sky(32, dark_sky)

    def test_upgrade(self):
        nside = 64
        dark_sky = sbp.dark_sky(nside=nside)

        self.assert_dark_sky(nside, dark_sky)

    def test_downgrade(self):
        nside = 16
        dark_sky = sbp.dark_sky(nside=nside)

        self.assert_dark_sky(nside, dark_sky)

    def test_downgrade_default(self):
        default_map_before_downgrade = sbp.dark_sky()

        sbp.dark_sky(nside=16)

        default_map_after_downgrade = sbp.dark_sky()

        for band in self.expected_bands:
            assert np.all(
                np.equal(
                    default_map_before_downgrade[band][np.isfinite(default_map_before_downgrade[band])],
                    default_map_after_downgrade[band][np.isfinite(default_map_after_downgrade[band])],
                )
            )

    def assert_dark_sky(self, nside, dark_sky):
        for band in self.expected_bands:
            assert band in dark_sky.dtype.names
            assert hp.npix2nside(len(dark_sky[band])) == nside

    @property
    def expected_bands(self):
        return ("u", "g", "r", "i", "z", "y")



__all__ = (
    "BaseFeature",
    "BaseSurveyFeature",
    "NObsCount",
    "NObsSurvey",
    "LastObservation",
    "LastsequenceObservation",
    "LastFilterChange",
    "NObservations",
    "CoaddedDepth",
    "LastObserved",
    "NObsNight",
    "PairInNight",
    "RotatorAngle",
    "NObservationsSeason",
    "NObsCountSeason",
    "NObservationsCurrentSeason",
    "LastNObsTimes",
    "SurveyInNight",
    "NoteInNight",
    "NoteLastObserved",
)

import warnings

import healpy as hp
import numpy as np
from scipy.stats import binned_statistic

from rubin_scheduler.scheduler import utils
from rubin_scheduler.scheduler.utils import IntRounded
from rubin_scheduler.skybrightness_pre import dark_sky
from rubin_scheduler.utils import _hpid2_ra_dec, calc_season, m5_flat_sed, survey_start_mjd


def send_unused_deprecation_warning(name):
    message = (
        f"The feature {name} is not in use by the current "
        "baseline scheduler and may be deprecated shortly. "
        "Please contact the rubin_scheduler maintainers if "
        "this is in use elsewhere."
    )
    warnings.warn(message, FutureWarning)


class BaseFeature:
    """The base class for features.
    This defines the standard API: a Feature should include
    a `self.feature` attribute, which could be a float, bool,
    or healpix size numpy array, or numpy masked array, and a
    __call__ method which returns `self.feature`.
    """

    def __init__(self, **kwargs):
        # self.feature should be a float, bool, or healpix size numpy
        # array, or numpy masked array
        self.feature = None

    def __call__(self):
        return self.feature


class BaseSurveyFeature(BaseFeature):
    """Track information relevant to the progress of a survey, using
    `self.feature` to hold this information.

    Features can track a single piece of information, or keep a map across
    the sky, or any other piece of information.

    Information in `self.feature` is updated via `add_observation` or
    `add_observation_array`.
    """

    def add_observations_array(self, observations_array, observations_hpid):
        """Update self.feature based on information in `observations_array`
        and `observations_hpid`.

        This is a method to more rapidly restore the feature to its
        expected state, by using an array of all previous observations
        (`observations_array`) instead looping over the individual
        observations, as in `self.add_observation`. The observations array
        allows rapid calculation of acceptable observations, without
        considering factors such as which healpix is relevant.

        The `observations_hpid` is a reorganized version of the
        `observations_array`, where the array representing each observation
        is extended to include `hpid` and recorded multiple times (for each
        healpixel). The `observations_hpid` allows rapid calculation of
        feature values which depend on hpid.

        Links between `observations_array` and `observations_hpid` can
        be made using `ID`.
        """
        print(self)
        raise NotImplementedError

    def add_observation(self, observation, indx=None, **kwargs):
        """Update self.feature based on information in `observation`.
        Observations should be ordered in monotonically increasing time.

        Parameters
        ----------
        observation : `np.array`, (1,N)
            Array of observation information, containing
            `mjd` for the time. See
            `rubin_scheduler.scheduler.utils.empty_observation`.
        indx : `list`-like of [`int`]
            The healpixel indices that the observation overlaps.
            See `rubin_scheduler.utils.HpInLsstFov`.
        """
        raise NotImplementedError


class SurveyInNight(BaseSurveyFeature):
    """Count appearances of `survey_str` within observation `note` in
    the current night; `survey_str` must be contained in `note`.

    Useful to keep track of how many times a survey has executed in a night.

    Parameters
    ----------
    survey_str : `str`, optional
        String to search for in observation `note`.
        String does not have to match `note` exactly,
        just be contained in `note`.
        Default of "" means any observation will match.
    """

    def __init__(self, survey_str=""):
        self.feature = 0
        self.survey_str = survey_str
        self.night = -100
        send_unused_deprecation_warning(self.__class__.__name__)

    def add_observation(self, observation, indx=None):
        if observation["night"] != self.night:
            self.night = observation["night"]
            self.feature = 0

        if self.survey_str in observation["note"]:
            self.feature += 1


class NoteInNight(BaseSurveyFeature):
    """Count appearances of any of `notes` in observation `note` in the
    current night; `note` must match one of `notes` exactly.

    Useful for keeping track of how many times a survey or other subset
    of visits has executed in a given night.

    Parameters
    ----------
    notes : `list` [`str`], optional
        List of strings to match against observation `note` values.
        The `note` must match one of the items in notes exactly.
        Default of None uses `[]` and will match any note.
    """

    def __init__(self, notes=None):
        self.feature = 0
        if notes is None:
            notes = []
        self.notes = notes
        self.current_night = -100

    def add_observations_array(self, observations_array, observations_hpid):
        # Identify the most recent night.
        if self.current_night != observations_array["night"][-1]:
            self.current_night = observations_array["night"][-1].copy()
            self.feature = 0
        # Identify observations within this night.
        indx = np.where(observations_array["night"] == observations_array["night"][-1])[0]
        # Count observations that match notes.
        for ind in indx:
            if observations_array["note"][ind] in self.notes:
                self.feature += 1

    def add_observation(self, observation, indx=None):
        if self.current_night != observation["night"]:
            self.current_night = observation["night"].copy()
            self.feature = 0
        if observation["note"] in self.notes:
            self.feature += 1


class NObsCount(BaseSurveyFeature):
    """Count the number of observations.
    Total number, not tracked over sky

    Parameters
    ----------
    filtername : `str` (None)
        The filter to count (if None, all filters counted)

    """

    def __init__(self, filtername=None, tag=None):
        self.feature = 0
        self.filtername = filtername
        # 'tag' is used in GoalStrictFilterBasisFunction
        self.tag = tag
        if self.tag is not None:
            warnings.warn(
                "Tag is not a supported element"
                "of the `observation` and this aspect of "
                "the feature will be "
                "deprecated in 2 minor releases.",
                DeprecationWarning,
                stack_level=2,
            )

    def add_observations_array(self, observations_array, observations_hpid):
        if self.filtername is None:
            self.feature += np.size(observations_array)
        else:
            in_filt = np.where(observations_array["filter"] == self.filtername)[0]
            self.feature += np.size(in_filt)

    def add_observation(self, observation, indx=None):
        if (self.filtername is None) and (self.tag is None):
            # Track all observations
            self.feature += 1
        elif (
            (self.filtername is not None)
            and (self.tag is None)
            and (observation["filter"][0] in self.filtername)
        ):
            # Track all observations on a specified filter
            self.feature += 1
        elif (self.filtername is None) and (self.tag is not None) and (observation["tag"][0] in self.tag):
            # Track all observations on a specified tag
            self.feature += 1
        elif (
            (self.filtername is None)
            and (self.tag is not None)
            and
            # Track all observations on a specified filter on a specified tag
            (observation["filter"][0] in self.filtername)
            and (observation["tag"][0] in self.tag)
        ):
            self.feature += 1


class NObsCountSeason(BaseSurveyFeature):
    """Count the number of observations in a season.

    Parameters
    ----------
    filtername : `str` (None)
        The filter to count (if None, all filters counted)

    Notes
    -----
    Uses `season_calc` to calculate season value.

    Seems unused - added deprecation warning.
    """

    def __init__(
        self,
        season,
        nside=None,
        filtername=None,
        tag=None,
        season_modulo=2,
        offset=None,
        max_season=None,
        season_length=365.25,
    ):
        self.feature = 0
        self.filtername = filtername
        self.tag = tag
        self.season = season
        self.season_modulo = season_modulo
        if offset is None:
            self.offset = np.zeros(hp.nside2npix(nside), dtype=int)
        else:
            self.offset = offset
        self.max_season = max_season
        self.season_length = season_length
        send_unused_deprecation_warning(self.__class__.__name__)

    def add_observation(self, observation, indx=None):
        season = utils.season_calc(
            observation["night"],
            modulo=self.season_modulo,
            offset=self.offset[indx],
            max_season=self.max_season,
            season_length=self.season_length,
        )
        if self.season in season:
            if (self.filtername is None) and (self.tag is None):
                # Track all observations
                self.feature += 1
            elif (
                (self.filtername is not None)
                and (self.tag is None)
                and (observation["filter"][0] in self.filtername)
            ):
                # Track all observations on a specified filter
                self.feature += 1
            elif (self.filtername is None) and (self.tag is not None) and (observation["tag"][0] in self.tag):
                # Track all observations on a specified tag
                self.feature += 1
            elif (
                (self.filtername is None)
                and (self.tag is not None)
                and
                # Track all observations on a specified filter on a
                # specified tag
                (observation["filter"][0] in self.filtername)
                and (observation["tag"][0] in self.tag)
            ):
                self.feature += 1


class NObsSurvey(BaseSurveyFeature):
    """Count the number of observations, whole sky (not per pixel).

    Parameters
    ----------
    note : `str` (None)
        Only count observations that contain str in their note field
    """

    def __init__(self, note=None):
        self.feature = 0
        self.note = note

    def add_observation(self, observation, indx=None):
        # Track all observations
        if self.note is None:
            self.feature += 1
        else:
            if self.note in observation["note"]:
                self.feature += 1


class LastObservation(BaseSurveyFeature):
    """Track the last observation.
    Useful if you want to see when the
    last time a survey took an observation.

    Parameters
    ----------
    survey_name : `str` (None)
        Only records if the survey name matches (or survey_name set to None)
    """

    def __init__(self, survey_name=None):
        # Will this work for observations read from a database???
        # "note" is definitely NOT guaranteed to match the survey_name.
        self.survey_name = survey_name
        # Start out with an empty observation
        self.feature = utils.empty_observation()

    def add_observations_array(self, observations_array, observations_hpid):
        if self.survey_name is not None:
            good = np.where(observations_array["note"] == self.survey_name)[0]
            if np.size(good) < 0:
                self.feature = observations_array[good[-1]]
        else:
            if len(observations_array) > 0:
                self.feature = observations_array[-1]

    def add_observation(self, observation, indx=None):
        if self.survey_name is not None:
            if self.survey_name in observation["note"]:
                self.feature = observation
        else:
            self.feature = observation


class LastsequenceObservation(BaseSurveyFeature):
    """When was the last observation"""

    def __init__(self, sequence_ids=""):
        self.sequence_ids = sequence_ids  # The ids of all sequence
        # observations...
        # Start out with an empty observation
        self.feature = utils.empty_observation()

    def add_observation(self, observation, indx=None):
        if observation["survey_id"] in self.sequence_ids:
            self.feature = observation


class LastFilterChange(BaseSurveyFeature):
    """Record when the filter last changed."""

    def __init__(self):
        self.feature = {"mjd": 0.0, "previous_filter": None, "current_filter": None}

    def add_observation(self, observation, indx=None):
        if self.feature["current_filter"] is None:
            self.feature["mjd"] = observation["mjd"][0]
            self.feature["previous_filter"] = None
            self.feature["current_filter"] = observation["filter"][0]
        elif observation["filter"][0] != self.feature["current_filter"]:
            self.feature["mjd"] = observation["mjd"][0]
            self.feature["previous_filter"] = self.feature["current_filter"]
            self.feature["current_filter"] = observation["filter"][0]


class NObservations(BaseSurveyFeature):
    """
    Track the number of observations that have been made across the sky.

    Parameters
    ----------
    filtername : `str` ('r')
        String or list that has all the filters that can count.
    nside : `int` (32)
        The nside of the healpixel map to use

    """

    def __init__(self, filtername=None, nside=None, survey_name=None):
        if nside is None:
            nside = utils.set_default_nside()

        self.feature = np.zeros(hp.nside2npix(nside), dtype=float)
        self.filtername = filtername
        self.survey_name = survey_name
        self.bins = np.arange(hp.nside2npix(nside) + 1) - 0.5

    def add_observations_array(self, observations_array, observations_hpid):
        valid_indx = np.ones(observations_hpid.size, dtype=bool)
        if self.filtername is not None:
            valid_indx[np.where(observations_hpid["filter"] != self.filtername)[0]] = False
        if self.survey_name is not None:
            tmp = [name in self.survey_name for name in observations_hpid["note"]]
            valid_indx = valid_indx * np.array(tmp)
        data = observations_hpid[valid_indx]
        if np.size(data) > 0:
            result, _be, _bn = binned_statistic(
                data["hpid"], np.ones(data.size), statistic=np.sum, bins=self.bins
            )
            self.feature += result

    def add_observation(self, observation, indx=None):
        if self.filtername is None or observation["filter"][0] in self.filtername:
            if self.survey_name is None or observation["note"] in self.survey_name:
                self.feature[indx] += 1


class NObservationsSeason(BaseSurveyFeature):
    """
    Track the number of observations that have been made across sky

    Parameters
    ----------
    season : `int`
        Only count observations in this season (year).
    filtername : `str` ('r')
        String or list that has all the filters that can count.
    nside : `int` (32)
        The nside of the healpixel map to use
    offset : `int` (0)
        The offset to use when computing the season (days)
    modulo : `int` (None)
        How to mod the years when computing season

    Notes
    -----
    Uses `season_calc` to calculate season value.
    """

    def __init__(
        self,
        season,
        filtername=None,
        nside=None,
        offset=0,
        modulo=None,
        max_season=None,
        season_length=365.25,
    ):
        if offset is None:
            offset = np.zeros(hp.nside2npix(nside), dtype=int)
        if nside is None:
            nside = utils.set_default_nside()

        self.feature = np.zeros(hp.nside2npix(nside), dtype=float)
        self.filtername = filtername
        ## How does this work if the default is 0 -- in add_observation
        # an index is referenced for offset, so the default should fail
        self.offset = offset
        self.modulo = modulo
        self.season = season
        self.max_season = max_season
        self.season_length = season_length
        send_unused_deprecation_warning(self.__class__.__name__)

    def add_observation(self, observation, indx=None):
        # How does this work if indx is None -- self.offset[indx] should fail
        observation_season = utils.season_calc(
            observation["night"],
            offset=self.offset[indx],
            modulo=self.modulo,
            max_season=self.max_season,
            season_length=self.season_length,
        )
        if self.season in observation_season:
            if self.filtername is None or observation["filter"][0] in self.filtername:
                self.feature[indx] += 1


class LargestN:
    def __init__(self, n):
        self.n = n

    def __call__(self, in_arr):
        if np.size(in_arr) < self.n:
            return -1
        result = in_arr[-self.n]
        return result


class LastNObsTimes(BaseSurveyFeature):
    """Record the last three observations for each healpixel"""

    def __init__(self, filtername=None, n_obs=3, nside=None):
        self.filtername = filtername
        self.n_obs = n_obs
        if nside is None:
            nside = utils.set_default_nside()
        self.feature = np.zeros((n_obs, hp.nside2npix(nside)), dtype=float)
        self.bins = np.arange(hp.nside2npix(nside) + 1) - 0.5

    def add_observations_array(self, observations_array, observations_hpid):
        # Assumes we're already sorted on mjd
        valid_indx = np.ones(observations_hpid.size, dtype=bool)
        if self.filtername is not None:
            valid_indx[np.where(observations_hpid["filter"] != self.filtername)[0]] = False
        data = observations_hpid[valid_indx]

        if np.size(data) > 0:
            for i in range(1, self.n_obs + 1):
                func = LargestN(i)
                result, _be, _bn = binned_statistic(data["hpid"], data["mjd"], statistic=func, bins=self.bins)
                # some_vals = np.where(np.sum(result, axis=1) > 0)[0]
                self.feature[-i, :] = result

    def add_observation(self, observation, indx=None):
        if self.filtername is None or observation["filter"][0] in self.filtername:
            self.feature[0:-1, indx] = self.feature[1:, indx]
            self.feature[-1, indx] = observation["mjd"]


class NObservationsCurrentSeason(BaseSurveyFeature):
    """Count the observations at each healpix, taken in the most
    recent season, that meet filter, seeing and m5 criteria.

    Useful for ensuring that "good quality" observations are acquired
    in each season at each point in the survey footprint.

    Parameters
    ----------
    filtername : `str`, optional
        If None (default) count observations in any filter.
        Otherwise, only count observations in the specified filter.
    nside : `int`, optional
        If None (default), use default nside for scheduler.
        Otherwise, set nside for the map to nside.
    seeing_fwhm_max : `float`, optional
        If None (default), count observations up to any seeing value.
        Otherwise, only count observations with better seeing (`FWHMeff`)
        than `seeing_fwhm_max`. In arcseconds.
    m5_penalty_max : `float`, optional
        If None (default), count observations with any m5 value.
        Otherwise, only count observations within this value of the
        dark sky map at this pixel.
        Only relevant if filtername is not None.
    mjd_start : `float`, optional
        If None, uses default survey_start_mjd for the start of the survey.
        This defines the starting year for counting seasons, so should
        be the start of the survey.
    """

    def __init__(
        self,
        filtername=None,
        nside=None,
        seeing_fwhm_max=None,
        m5_penalty_max=None,
        mjd_start=None,
    ):
        if nside is None:
            nside = utils.set_default_nside()
        self.nside = nside
        self.seeing_fwhm_max = seeing_fwhm_max
        self.filtername = filtername
        self.m5_penalty_max = m5_penalty_max
        # If filtername not set, then we can't set m5_penalty_max.
        if self.filtername is None and self.m5_penalty_max is not None:
            warnings.warn("To use m5_penalty_max, filtername must be set. Disregarding m5_penalty_max.")
            self.m5_penalty_max = None

        # Get the dark sky map
        if self.filtername is not None:
            self.dark_map = dark_sky(nside)[filtername]
        self.ones = np.ones(hp.nside2npix(self.nside))

        if mjd_start is None:
            mjd_start = survey_start_mjd()
        self.mjd_start = mjd_start

        # Set up feature values - this includes the count in a given season
        # Find the healpixels for each point on the sky
        self.ra, self.dec = _hpid2_ra_dec(nside, np.arange(hp.nside2npix(nside)))
        self.ra_deg = np.degrees(self.ra)

        self.mjd = mjd_start
        # Set up season_map. This should be the same as conditions.season_map
        # but the important thing is that mjd_start matches.
        self.season_map = calc_season(self.ra_deg, [self.mjd], self.mjd_start).flatten()
        self.season = np.floor(self.season_map)

        self.feature = np.zeros(hp.nside2npix(nside), dtype=float)
        # Set bins for add_observations_array
        self.bins = np.arange(hp.nside2npix(nside) + 1) - 0.5

    def season_update(self, observation=None, conditions=None):
        """Update the season_map to the current time.

        This assumes time increases monotonically in the conditions or
        observations objects passed as arguments.
        Using the 'mjd' from the conditions, where parts of the sky
        have changed season (increasing in the `self.season_map` + 1)
        the `self.feature` is cleared, in order to restart counting
        observations in the new season.

        Parameters
        ----------
        observation : `np.array`, (1,N)
            Array of observation information, containing
            `mjd` for the time. See
            `rubin_scheduler.scheduler.utils.empty_observation`.
        conditions : `rubin_scheduler.scheduler.Conditions`, optional
            A conditions object, containing `mjd`.

        Notes
        -----
        One of observations or conditions must be passed, but either
        are options so this can be used with add_observation.
        Most of the time, the updates will come from `conditions`.
        """
        mjd_now = None
        mjd_from = None
        if observation is not None:
            mjd_now = observation["mjd"]
            mjd_from = "observation"
        # Prefer to use Conditions for the time.
        if conditions is not None:
            mjd_now = conditions.mjd
            mjd_from = "conditions"
        if mjd_now is None:
            warnings.warn(
                "Expected either a conditions or observations object. Not updating season_map.",
                UserWarning,
            )
            return
        if isinstance(mjd_now, np.ndarray) or isinstance(mjd_now, list):
            mjd_now = mjd_now[0]
        # Check that time doesn't go backwards.
        # But only bother to warn if it's coming from conditions.
        # Observations may look like they go backwards.
        if mjd_now < self.mjd:
            if mjd_from == "conditions":
                warnings.warn(
                    f"Time must flow forwards to track the "
                    f"feature in {self.__class__.__name__}."
                    f"Not updating season_map.",
                    UserWarning,
                )
            # But just return without update in either case.
            return
        # Calculate updated season values.
        updated_season = np.floor(self.season_map + (mjd_now - self.mjd_start) / 365.25)
        # Clear feature where season increased by 1 (note 'floor' above).
        self.feature[np.where(updated_season > self.season)] = 0
        # Update to new time and season.
        self.mjd = mjd_now
        self.season = updated_season

    def add_observations_array(self, observations_array, observations_hpid):
        # Update self.season to the most recent observation time
        self.season_update(observation=observations_array[-1])

        # Start assuming 'check' is all True, for all observations.
        check = np.ones(observations_array.size, dtype=bool)

        # Rule out the observations with seeing fwhmeff > limit
        if self.seeing_fwhm_max is not None:
            check[np.where(observations_array["FWHMeff"] > self.seeing_fwhm_max)] = False

        # Rule out observations which are not in the desired filter
        if self.filtername is not None:
            check[np.where(observations_array["filter"] != self.filtername)] = False

        # Convert the "check" array on observations_array into a
        # "check" array on the hpids so we can evaluate healpix-level items
        check_hpid_indxs = np.in1d(observations_hpid["ID"], observations_array["ID"][check])
        # Set up a new valid/not valid flag, now on observations_hpid
        check_hp = np.zeros(len(observations_hpid), dtype=bool)
        # Set all observations which passed simpler tests above to True
        check_hp[check_hpid_indxs] = True

        hpids = observations_hpid["hpid"]

        # This could be done once per observation, but to make life
        # easier for index matching, let's just calculate season
        # based on the hpid array.
        # We only want to count observations that would fall within the
        # current season, so calculate the season for each obs.
        seasons = np.floor(
            calc_season(
                np.degrees(observations_hpid["RA"]),
                observations_hpid["mjd"],
                self.mjd_start,
                calc_diagonal=True,
            )
        )
        season_change = self.season[hpids] - seasons
        check_hp = np.where(season_change != 0, False, check_hp)
        # And check for m5_penalty_map
        if self.m5_penalty_max is not None:
            penalty = self.dark_map[hpids] - observations_hpid["fivesigmadepth"]
            check_hp = np.where(penalty > self.m5_penalty_max, False, check_hp)

        # Bin up the observations per hpid and add them to the feature.
        result, _be, _bn = binned_statistic(
            observations_hpid["hpid"][check_hp],
            observations_hpid["hpid"][check_hp],
            bins=self.bins,
            statistic=np.size,
        )
        # Add the resulting observations to feature.
        self.feature += result

    def add_observation(self, observation, indx):
        # Update the season map to the current time.
        self.season_update(observation=observation)
        # Check if *this* observation is in the current season
        # (This means we could add observations from the past, but
        # would count it against the current season).
        season_obs = np.floor(self.season_map[indx] + (observation["mjd"] - self.mjd_start) / 365.25)
        this_season_indx = np.array(indx)[np.where(season_obs == self.season[indx])]

        # Check if seeing is good enough.
        if self.seeing_fwhm_max is not None:
            check = observation["FWHMeff"] <= self.seeing_fwhm_max
        else:
            check = True

        # Check if observation is in the right filter
        if self.filtername is not None and check:
            if observation["filter"] != self.filtername:
                check = False
            else:
                # Check if observation in correct filter and deep enough.
                if self.m5_penalty_max is not None:
                    penalty = self.dark_map[this_season_indx] - observation["fivesigmadepth"]
                    this_season_indx = this_season_indx[np.where(penalty <= self.m5_penalty_max)]

        if check:
            self.feature[this_season_indx] += 1


class CoaddedDepth(BaseSurveyFeature):
    """Track the co-added depth that has been reached across the sky

    Parameters
    ----------
    fwh_meff_limit : `float` (100)
        The effective FWHM of the seeing (arcsecond).
        Images will only be added to the coadded depth if the observation
        FWHM is less than or equal to the limit.  Default 100.
    """

    def __init__(self, filtername="r", nside=None, fwhm_eff_limit=100.0):
        if nside is None:
            nside = utils.set_default_nside()
        self.filtername = filtername
        self.fwhm_eff_limit = IntRounded(fwhm_eff_limit)
        # Starting at limiting mag of zero should be fine.
        self.feature = np.zeros(hp.nside2npix(nside), dtype=float)

    def add_observation(self, observation, indx=None):
        if observation["filter"] == self.filtername:
            if IntRounded(observation["FWHMeff"]) <= self.fwhm_eff_limit:
                m5 = m5_flat_sed(
                    observation["filter"],
                    observation["skybrightness"],
                    observation["FWHMeff"],
                    observation["exptime"],
                    observation["airmass"],
                )

                self.feature[indx] = 1.25 * np.log10(10.0 ** (0.8 * self.feature[indx]) + 10.0 ** (0.8 * m5))


class LastObserved(BaseSurveyFeature):
    """
    Track when a pixel was last observed.
    Assumes observations are added in chronological order.
    """

    def __init__(self, filtername="r", nside=None, fill=np.nan):
        if nside is None:
            nside = utils.set_default_nside()

        self.filtername = filtername
        self.feature = np.zeros(hp.nside2npix(nside), dtype=float) + fill
        self.bins = np.arange(hp.nside2npix(nside) + 1) - 0.5

    def add_observations_array(self, observations_array, observations_hpid):
        # Assumes we're already sorted on mjd
        valid_indx = np.ones(observations_hpid.size, dtype=bool)
        if self.filtername is not None:
            valid_indx[np.where(observations_hpid["filter"] != self.filtername)[0]] = False
        data = observations_hpid[valid_indx]

        if np.size(data) > 0:
            result, _be, _bn = binned_statistic(data["hpid"], data["mjd"], statistic=np.max, bins=self.bins)
            good = np.where(result > 0)
            self.feature[good] = result[good]

    def add_observation(self, observation, indx=None):
        if self.filtername is None:
            self.feature[indx] = observation["mjd"]
        elif observation["filter"][0] in self.filtername:
            self.feature[indx] = observation["mjd"]


class NoteLastObserved(BaseSurveyFeature):
    """Track the last time an observation with a particular `note` field was
    made.

    Parameters
    ----------
    note : `str`
        Substring to match an observation note field to keep track of.
    """

    def __init__(self, note, filtername=None):
        self.note = note
        self.filtername = filtername
        self.feature = None

    def add_observation(self, observation, indx=None):
        if self.note in observation["note"] and (
            self.filtername is None or self.filtername == observation["filter"]
        ):
            self.feature = observation["mjd"]


class NObsNight(BaseSurveyFeature):
    """
    Track how many times something has been observed in a night
    (Note, even if there are two, it might not be a good pair.)

    Parameters
    ----------
    filtername : `str` ('r')
        Filter to track.
    nside : `int` (32)
        Scale of the healpix map

    """

    def __init__(self, filtername="r", nside=None):
        if nside is None:
            nside = utils.set_default_nside()

        self.filtername = filtername
        self.feature = np.zeros(hp.nside2npix(nside), dtype=int)
        self.night = None

    def add_observation(self, observation, indx=None):
        if observation["night"] != self.night:
            self.feature *= 0
            self.night = observation["night"]
        if (self.filtername == "") | (self.filtername is None):
            self.feature[indx] += 1
        elif observation["filter"][0] in self.filtername:
            self.feature[indx] += 1


class PairInNight(BaseSurveyFeature):
    """
    Track how many pairs have been observed within a night

    Parameters
    ----------
    gap_min : `float` (25.)
        The minimum time gap to consider a successful pair in minutes
    gap_max : `float` (45.)
        The maximum time gap to consider a successful pair (minutes)
    """

    def __init__(self, filtername="r", nside=None, gap_min=25.0, gap_max=45.0):
        if nside is None:
            nside = utils.set_default_nside()

        self.filtername = filtername
        self.feature = np.zeros(hp.nside2npix(nside), dtype=float)
        self.indx = np.arange(self.feature.size)
        self.last_observed = LastObserved(filtername=filtername)
        self.gap_min = IntRounded(gap_min / (24.0 * 60))  # Days
        self.gap_max = IntRounded(gap_max / (24.0 * 60))  # Days
        self.night = 0
        # Need to keep a full record of times and healpixels observed in
        # a night.
        self.mjd_log = []
        self.hpid_log = []

    def add_observations_array(self, observations_array, observations_hpid):
        # ok, let's just find the largest night and toss all those in one
        # at a time
        most_recent_night = np.where(observations_hpid["night"] == np.max(observations_hpid["night"]))[0]
        obs_hpid = observations_hpid[most_recent_night]
        uid = np.unique(obs_hpid["ID"])
        for ind_id in uid:
            # maybe a faster searchsorted way to do this, but it'll work
            # for now
            good = np.where(obs_hpid["ID"] == ind_id)[0]
            self.add_observation(observations_hpid[good][0], observations_hpid[good]["hpid"])

    def add_observation(self, observation, indx=None):
        if self.filtername is None:
            infilt = True
        else:
            infilt = observation["filter"][0] in self.filtername
        if infilt:
            if indx is None:
                indx = self.indx
            # Clear values if on a new night
            if self.night != observation["night"]:
                self.feature *= 0.0
                self.night = observation["night"]
                self.mjd_log = []
                self.hpid_log = []

            # Look for the mjds that could possibly pair with observation
            mjd_diff = IntRounded(observation["mjd"] - np.array(self.mjd_log))
            # normally would use np.searchsorted, but need to use IntRounded
            # to be sure we are cross-platform repeatable.
            in_range_indx = np.where((mjd_diff > self.gap_min) & (mjd_diff < self.gap_max))[0]
            # Now check if any of the healpixels taken in the time gap
            # match the healpixels of the observation.
            if in_range_indx.size > 0:
                left = in_range_indx.min()
                right = in_range_indx.max() + 1
                matches = np.in1d(indx, self.hpid_log[left:right])
                self.feature[np.array(indx)[matches]] += 1

            # record the mjds and healpixels that were observed
            self.mjd_log.extend([np.max(observation["mjd"])] * np.size(indx))
            self.hpid_log.extend(list(indx))


class RotatorAngle(BaseSurveyFeature):
    """
    Track what rotation angles things are observed with.
    XXX-under construction
    """

    def __init__(self, filtername="r", binsize=10.0, nside=None):
        """"""
        if nside is None:
            nside = utils.set_default_nside()

        self.filtername = filtername
        # Actually keep a histogram at each healpixel
        self.feature = np.zeros((hp.nside2npix(nside), 360.0 / binsize), dtype=float)
        self.bins = np.arange(0, 360 + binsize, binsize)

    def add_observation(self, observation, indx=None):
        if observation["filter"][0] == self.filtername:
            # I think this is how to broadcast things properly.
            self.feature[indx, :] += np.histogram(observation.rotSkyPos, bins=self.bins)[0]



import unittest

import healpy as hp
import numpy as np
from astropy import units as u
from astropy.coordinates import AltAz, EarthLocation, SkyCoord
from astropy.time import Time

import rubin_scheduler.utils as utils


def astropy_ra_dec_to_alt_az(ra, dec, mjd, lat_deg, longitude_deg, height=100):
    observing_location = EarthLocation(lat=lat_deg * u.deg, lon=longitude_deg * u.deg, height=height * u.m)
    observing_time = Time(mjd, format="mjd")
    aa = AltAz(location=observing_location, obstime=observing_time)

    coord = SkyCoord(ra * u.deg, dec * u.deg)
    alt_az = coord.transform_to(aa)
    return alt_az.alt.deg, alt_az.az.deg


class ApproxCoordTests(unittest.TestCase):
    """
    Test the fast approximate ra,dec to alt,az transforms
    """

    def test_degrees(self):
        nside = 16
        hpids = np.arange(hp.nside2npix(nside))
        ra, dec = utils.hpid2_ra_dec(nside, hpids)
        mjd = 59852.0
        site = utils.Site("LSST")

        alt1, az1 = astropy_ra_dec_to_alt_az(ra, dec, mjd, site.latitude, site.longitude)
        alt2, az2 = utils.approx_ra_dec2_alt_az(ra, dec, site.latitude, site.longitude, mjd)

        # Check that the fast is similar to the more precice transform
        tol = 2  # Degrees
        tol_mean = 1.0
        separations = utils.angular_separation(az1, alt1, az2, alt2)
        self.assertLess(np.max(separations), tol)
        self.assertLess(np.mean(separations), tol_mean)

        # Check that the fast can nearly round-trip
        ra_back, dec_back = utils.approx_alt_az2_ra_dec(alt2, az2, site.latitude, site.longitude, mjd)
        separations = utils.angular_separation(ra, dec, ra_back, dec_back)
        self.assertLess(np.max(separations), tol)
        self.assertLess(np.mean(separations), tol_mean)

    def test_rad(self):
        nside = 16
        hpids = np.arange(hp.nside2npix(nside))
        ra, dec = utils._hpid2_ra_dec(nside, hpids)
        mjd = 59852.0
        site = utils.Site("LSST")

        alt1, az1 = astropy_ra_dec_to_alt_az(
            np.degrees(ra), np.degrees(dec), mjd, site.latitude, site.longitude
        )
        alt1 = np.radians(alt1)
        az1 = np.radians(az1)

        alt2, az2 = utils._approx_ra_dec2_alt_az(ra, dec, site.latitude_rad, site.longitude_rad, mjd)

        # Check that the fast is similar to the more precice transform
        tol = np.radians(2)
        tol_mean = np.radians(1.0)
        separations = utils._angular_separation(az1, alt1, az2, alt2)

        self.assertLess(np.max(separations), tol)
        self.assertLess(np.mean(separations), tol_mean)

        # Check that the fast can nearly round-trip
        ra_back, dec_back = utils._approx_alt_az2_ra_dec(
            alt2, az2, site.latitude_rad, site.longitude_rad, mjd
        )
        separations = utils._angular_separation(ra, dec, ra_back, dec_back)
        self.assertLess(np.max(separations), tol)
        self.assertLess(np.mean(separations), tol_mean)


if __name__ == "__main__":
    unittest.main()



__all__ = (
    "FilterLoadedBasisFunction",
    "OnceInNightBasisFunction",
    "SunAltHighLimitBasisFunction",
    "TimeToTwilightBasisFunction",
    "NotTwilightBasisFunction",
    "AfterEveningTwiBasisFunction",
    "ForceDelayBasisFunction",
    "HourAngleLimitBasisFunction",
    "MoonDownBasisFunction",
    "FractionOfObsBasisFunction",
    "CloudedOutBasisFunction",
    "RisingMoreBasisFunction",
    "SoftDelayBasisFunction",
    "LookAheadDdfBasisFunction",
    "SunAltLimitBasisFunction",
    "TimeInTwilightBasisFunction",
    "NightModuloBasisFunction",
    "EndOfEveningBasisFunction",
    "TimeToScheduledBasisFunction",
    "LimitObsPnightBasisFunction",
    "SunHighLimitBasisFunction",
)

import numpy as np

from rubin_scheduler.scheduler import features
from rubin_scheduler.scheduler.basis_functions import BaseBasisFunction
from rubin_scheduler.scheduler.utils import IntRounded


class FilterLoadedBasisFunction(BaseBasisFunction):
    """Check that the filter(s) needed are loaded.

    Are the filters in `filternames` loaded and available?

    Parameters
    ----------
    filternames : `str` or `list` [ `str` ]
        The filternames that need to be mounted to execute.
    """

    def __init__(self, filternames="r"):
        super(FilterLoadedBasisFunction, self).__init__()
        if not isinstance(filternames, list):
            filternames = [filternames]
        self.filternames = filternames

    def check_feasibility(self, conditions):
        for filtername in self.filternames:
            result = filtername in conditions.mounted_filters
            if result is False:
                return result
        return result


class SunHighLimitBasisFunction(BaseBasisFunction):
    """Only execute if the sun is higher than `sun_alt_limit`,
    the current time is within `time_to_12deg` of -12 degree twilight,
    and there is at least `time_remaining` time left before -12 degree
    twilight.

    Is the current time and sun altitude close to twilight, but not too close?

    Parameters
    ----------
    sun_alt_limit : `float`
        The sun altitude limit (degrees). Sun must be higher than this
        at sunset to execute
    time_to_12deg : `float`
        How much time must be remaining before 12 degree twilight in
        the morning (minutes)
    time_remaining : `float`
        Minimum about of time that must be available before trying to
        execute (minutes)

    Notes
    -----
    This is primarily useful for surveys which must execute within close
    limits of -12 degree twilight.
    """

    def __init__(self, sun_alt_limit=-14.8, time_to_12deg=21.0, time_remaining=15.0):
        super(SunHighLimitBasisFunction, self).__init__()
        if time_to_12deg < time_remaining:
            raise ValueError(
                "time_to_12deg value of %f is less than time_remaining value of %f."
                % (time_to_12deg, time_remaining)
            )
        self.sun_alt_limit = np.radians(sun_alt_limit)
        self.time_to_12deg = time_to_12deg / 60.0 / 24.0
        self.time_remaining = time_remaining / 60.0 / 24.0

    def check_feasibility(self, conditions):
        result = False

        # If the sun is high, it's ok to execute
        if conditions.sun_alt > self.sun_alt_limit:
            result = True
        time_left = conditions.sun_n12_rising - conditions.mjd
        if time_left < self.time_to_12deg:
            result = True
        if time_left < self.time_remaining:
            result = False
        return result


class OnceInNightBasisFunction(BaseBasisFunction):
    """Stop observing if something has been executed already in the night

    Parameters
    ----------
    notes : `list` [ `str` ]
        A list of str to check if any observations with a matching note exist.
    """

    def __init__(self, notes=[]):
        super(OnceInNightBasisFunction, self).__init__()
        self.survey_features["note_in_night"] = features.NoteInNight(notes=notes)

    def check_feasibility(self, conditions):
        result = True
        if self.survey_features["note_in_night"].feature > 0:
            result = False
        return result


class SunAltHighLimitBasisFunction(BaseBasisFunction):
    """Don't observe unless the sun is above some limit

    Parameters
    ----------
    alt_limit : float (-15)
        Sun must be above alt_limit (degrees). Default -15.
    """

    def __init__(self, alt_limit=-15.0):
        super(SunAltHighLimitBasisFunction, self).__init__()
        self.alt_limit = np.radians(alt_limit)

    def check_feasibility(self, conditions):
        result = True
        if conditions.sun_alt < self.alt_limit:
            result = False
        return result


class LimitObsPnightBasisFunction(BaseBasisFunction):
    """"""

    def __init__(self, survey_str="", nlimit=100.0):
        super(LimitObsPnightBasisFunction, self).__init__()
        self.nlimit = nlimit
        self.survey_features["N_in_night"] = features.Survey_in_night(survey_str=survey_str)

    def check_feasibility(self, conditions):
        if self.survey_features["N_in_night"].feature >= self.nlimit:
            return False
        else:
            return True


class NightModuloBasisFunction(BaseBasisFunction):
    """Only return true on certain nights"""

    def __init__(self, pattern=None):
        super(NightModuloBasisFunction, self).__init__()
        if pattern is None:
            pattern = [True, False]
        self.pattern = pattern
        self.mod_val = len(self.pattern)

    def check_feasibility(self, conditions):
        indx = int(conditions.night % self.mod_val)
        result = self.pattern[indx]
        return result


class TimeInTwilightBasisFunction(BaseBasisFunction):
    """Make sure there is some time left in twilight.

    Parameters
    ----------
    time_needed : float (5)
        The time needed remaining in twilight (minutes)
    """

    def __init__(self, time_needed=5.0):
        super(TimeInTwilightBasisFunction, self).__init__()
        self.time_needed = time_needed / 60.0 / 24.0  # To days

    def check_feasibility(self, conditions):
        result = False
        time1 = conditions.sun_n18_setting - conditions.mjd
        time2 = conditions.sun_n12_rising - conditions.mjd

        if time1 > self.time_needed:
            result = True
        else:
            if conditions.sunAlt > np.radians(-18.0):
                if time2 > self.time_needed:
                    result = True
        return result


class AfterEveningTwiBasisFunction(BaseBasisFunction):
    """Only execute right after evening twilight"""

    def __init__(self, time_after=30.0, alt_limit=18):
        super(AfterEveningTwiBasisFunction, self).__init__()
        self.time_after = IntRounded(time_after / 60.0 / 24.0)
        self.alt_limit = str(alt_limit)

    def check_feasibility(self, conditions):
        available_time = conditions.mjd - getattr(conditions, "sun_n" + self.alt_limit + "_setting")
        result = IntRounded(available_time) < self.time_after
        return result


class EndOfEveningBasisFunction(BaseBasisFunction):
    """Only let observations happen in a limited time before twilight"""

    def __init__(self, time_remaining=30.0, alt_limit=18):
        super(EndOfEveningBasisFunction, self).__init__()
        self.time_remaining = IntRounded(time_remaining / 60.0 / 24.0)
        self.alt_limit = str(alt_limit)

    def check_feasibility(self, conditions):
        available_time = getattr(conditions, "sun_n" + self.alt_limit + "_rising") - conditions.mjd
        result = IntRounded(available_time) < self.time_remaining
        return result


class TimeToTwilightBasisFunction(BaseBasisFunction):
    """Make sure there is enough time before twilight. Useful
    if you want to check before starting a long sequence of observations.

    Parameters
    ----------
    time_needed : float (30.)
        The time needed to run a survey (mintues).
    alt_limit : int (18)
        The sun altitude limit to use. Must be 12 or 18
    """

    def __init__(self, time_needed=30.0, alt_limit=18):
        super(TimeToTwilightBasisFunction, self).__init__()
        self.time_needed = time_needed / 60.0 / 24.0  # To days
        self.alt_limit = str(alt_limit)

    def check_feasibility(self, conditions):
        available_time = getattr(conditions, "sun_n" + self.alt_limit + "_rising") - conditions.mjd
        result = available_time > self.time_needed
        return result


class TimeToScheduledBasisFunction(BaseBasisFunction):
    """Make sure there is enough time before next scheduled observation.
    Useful if you want to check before starting a long sequence of
    observations.

    Parameters
    ----------
    time_needed : float (30.)
        The time needed to run a survey (mintues).
    """

    def __init__(self, time_needed=30.0):
        super(TimeToScheduledBasisFunction, self).__init__()
        self.time_needed = time_needed / 60.0 / 24.0  # To days

    def check_feasibility(self, conditions):
        if len(conditions.scheduled_observations) == 0:
            return True

        available_time = np.min(conditions.scheduled_observations) - conditions.mjd
        result = available_time > self.time_needed
        return result


class NotTwilightBasisFunction(BaseBasisFunction):
    def __init__(self, sun_alt_limit=-18):
        """
        # Should be -18 or -12
        """
        self.sun_alt_limit = str(int(sun_alt_limit)).replace("-", "n")
        super(NotTwilightBasisFunction, self).__init__()

    def check_feasibility(self, conditions):
        result = True
        if conditions.mjd < getattr(conditions, "sun_" + self.sun_alt_limit + "_setting"):
            result = False
        if conditions.mjd > getattr(conditions, "sun_" + self.sun_alt_limit + "_rising"):
            result = False
        return result


class ForceDelayBasisFunction(BaseBasisFunction):
    """Keep a survey from executing to rapidly.

    Parameters
    ----------
    days_delay : float (2)
        The number of days to force a gap on.
    """

    def __init__(self, days_delay=2.0, survey_name=None):
        super(ForceDelayBasisFunction, self).__init__()
        self.days_delay = days_delay
        self.survey_name = survey_name
        self.survey_features["last_obs_self"] = features.LastObservation(survey_name=self.survey_name)

    def check_feasibility(self, conditions):
        result = True
        if conditions.mjd - self.survey_features["last_obs_self"].feature["mjd"] < self.days_delay:
            result = False
        return result


class SoftDelayBasisFunction(BaseBasisFunction):
    """Like Force_delay, but go ahead and let things catch up if they fall
    far behind.

    Parameters
    ----------

    """

    def __init__(self, fractions=[0.000, 0.009, 0.017], delays=[0.0, 0.5, 1.5], survey_name=None):
        if len(fractions) != len(delays):
            raise ValueError("fractions and delays must be same length")
        super(SoftDelayBasisFunction, self).__init__()
        self.delays = delays
        self.survey_name = survey_name
        self.survey_features["last_obs_self"] = features.LastObservation(survey_name=self.survey_name)
        self.fractions = fractions
        self.survey_features["Ntot"] = features.NObsSurvey()
        self.survey_features["N_survey"] = features.NObsSurvey(note=self.survey_name)

    def check_feasibility(self, conditions):
        result = True
        current_ratio = self.survey_features["N_survey"].feature / self.survey_features["Ntot"].feature
        indx = np.searchsorted(self.fractions, current_ratio)
        if indx == len(self.fractions):
            indx -= 1
        delay = self.delays[indx]
        if conditions.mjd - self.survey_features["last_obs_self"].feature["mjd"] < delay:
            result = False
        return result


class HourAngleLimitBasisFunction(BaseBasisFunction):
    """Only execute a survey in limited hour angle ranges. Useful for
    limiting Deep Drilling Fields.

    Parameters
    ----------
    RA : float (0.)
        RA of the target (degrees).
    ha_limits : list of lists
        limits for what hour angles are acceptable (hours). e.g.,
        to give 4 hour window around RA=0, ha_limits=[[22,24], [0,2]]
    """

    def __init__(self, RA=0.0, ha_limits=None):
        super(HourAngleLimitBasisFunction, self).__init__()
        self.ra_hours = RA / 360.0 * 24.0
        self.ha_limits = np.array(ha_limits)

    def check_feasibility(self, conditions):
        target_ha = (conditions.lmst - self.ra_hours) % 24
        # Are we in any of the possible windows
        result = False
        for limit in self.ha_limits:
            lres = limit[0] <= target_ha < limit[1]
            result = result or lres

        return result


class MoonDownBasisFunction(BaseBasisFunction):
    """Demand the moon is down"""

    def check_feasibility(self, conditions):
        result = True
        if conditions.moon_alt > 0:
            result = False
        return result


class FractionOfObsBasisFunction(BaseBasisFunction):
    """Limit the fraction of all observations that can be labled a certain
    survey name. Useful for keeping DDFs from exceeding a given fraction
    of the total survey.

    Parameters
    ----------
    frac_total : float
        The fraction of total observations that can be of this survey
    survey_name : str
        The name of the survey
    """

    def __init__(self, frac_total, survey_name=""):
        super(FractionOfObsBasisFunction, self).__init__()
        self.survey_name = survey_name
        self.frac_total = frac_total
        self.survey_features["Ntot"] = features.NObsSurvey()
        self.survey_features["N_survey"] = features.NObsSurvey(note=self.survey_name)

    def check_feasibility(self, conditions):
        # If nothing has been observed, fine to go
        result = True
        if self.survey_features["Ntot"].feature == 0:
            return result
        ratio = self.survey_features["N_survey"].feature / self.survey_features["Ntot"].feature
        if ratio > self.frac_total:
            result = False
        return result


class LookAheadDdfBasisFunction(BaseBasisFunction):
    """Look into the future to decide if it's a good time to observe or block.

    Parameters
    ----------
    frac_total : float
        The fraction of total observations that can be of this survey
    aggressive_fraction : float
        If the fraction of observations drops below ths value, be more
        aggressive in scheduling.  e.g., do not wait for conditions to
        improve, execute as soon as possible.
    time_needed : float (30.)
        Estimate of the amount of time needed to execute DDF sequence
        (minutes).
    RA : float (0.)
        The RA of the DDF
    ha_limits : list of lists (None)
        limits for what hour angles are acceptable (hours). e.g.,
        to give 4 hour window around HA=0, ha_limits=[[22,24], [0,2]]
    survey_name : str ('')
        The name of the survey
    time_jump : float (44.)
        The amount of time to assume will jump ahead if another survey
        executes (minutes)
    sun_alt_limit : float (-18.)
        The limit to assume twilight starts (degrees)
    """

    def __init__(
        self,
        frac_total,
        aggressive_fraction,
        time_needed=30.0,
        RA=0.0,
        ha_limits=None,
        survey_name="",
        time_jump=44.0,
        sun_alt_limit=-18.0,
    ):
        super(LookAheadDdfBasisFunction, self).__init__()
        if aggressive_fraction > frac_total:
            raise ValueError("aggressive_fraction should be less than frac_total")
        self.survey_name = survey_name
        self.frac_total = frac_total
        self.ra_hours = RA / 360.0 * 24.0
        self.ha_limits = np.array(ha_limits)
        self.sun_alt_limit = str(int(sun_alt_limit)).replace("-", "n")
        self.time_jump = time_jump / 60.0 / 24.0  # To days
        self.time_needed = time_needed / 60.0 / 24.0  # To days
        self.aggressive_fraction = aggressive_fraction
        self.survey_features["Ntot"] = features.NObsSurvey()
        self.survey_features["N_survey"] = features.NObsSurvey(note=self.survey_name)

    def check_feasibility(self, conditions):
        result = True
        target_ha = (conditions.lmst - self.ra_hours) % 24
        ratio = self.survey_features["N_survey"].feature / self.survey_features["Ntot"].feature
        available_time = getattr(conditions, "sun_" + self.sun_alt_limit + "_rising") - conditions.mjd
        # If it's more that self.time_jump to hour angle zero
        # See if there will be enough time to twilight in the future
        if (IntRounded(target_ha) > IntRounded(12)) & (
            IntRounded(target_ha) < IntRounded(24.0 - self.time_jump)
        ):
            if IntRounded(available_time) > IntRounded(self.time_needed + self.time_jump):
                result = False
                # If we paused for better conditions, but the moon will
                # rise, turn things back on.
                if IntRounded(conditions.moon_alt) < IntRounded(0):
                    if IntRounded(conditions.moonrise) > IntRounded(conditions.mjd):
                        if IntRounded(conditions.moonrise - conditions.mjd) > IntRounded(self.time_jump):
                            result = True
        # If the moon is up and will set soon, pause
        if IntRounded(conditions.moon_alt) > IntRounded(0):
            time_after_moonset = (
                getattr(conditions, "sun_" + self.sun_alt_limit + "_rising") - conditions.moonset
            )
            if IntRounded(conditions.moonset) > IntRounded(self.time_jump):
                if IntRounded(time_after_moonset) > IntRounded(self.time_needed):
                    result = False

        # If the survey has fallen far behind, be agressive and observe
        # anytime it's up.
        if IntRounded(ratio) < IntRounded(self.aggressive_fraction):
            result = True
        return result


class CloudedOutBasisFunction(BaseBasisFunction):
    def __init__(self, cloud_limit=0.7):
        super(CloudedOutBasisFunction, self).__init__()
        self.cloud_limit = cloud_limit

    def check_feasibility(self, conditions):
        result = True
        if conditions.bulk_cloud > self.cloud_limit:
            result = False
        return result


class RisingMoreBasisFunction(BaseBasisFunction):
    """Say a spot is not available if it will rise substatially before
    twilight.

    Parameters
    ----------
    RA : float
        The RA of the point in the sky (degrees)
    pad : float
        When to start observations if there's plenty of time before
        twilight (minutes)
    """

    def __init__(self, RA, pad=30.0):
        super(RisingMoreBasisFunction, self).__init__()
        self.ra_hours = RA * 24 / 360.0
        self.pad = pad / 60.0  # To hours

    def check_feasibility(self, conditions):
        result = True
        hour_angle = conditions.lmst - self.ra_hours
        # If it's rising, and twilight is well beyond when it crosses
        # the meridian
        time_to_twi = (conditions.sun_n18_rising - conditions.mjd) * 24.0
        if (hour_angle < -self.pad) & (np.abs(hour_angle) < (time_to_twi - self.pad)):
            result = False
        return result


class SunAltLimitBasisFunction(BaseBasisFunction):
    """Don't try unless the sun is below some limit"""

    def __init__(self, alt_limit=-12.1):
        super(SunAltLimitBasisFunction, self).__init__()
        self.alt_limit = np.radians(alt_limit)

    def check_feasibility(self, conditions):
        result = True
        if conditions.sunAlt > self.alt_limit:
            result = False
        return result



__all__ = ("MyHTMLParser", "rs_download_sky")

import argparse
import os
from html.parser import HTMLParser

import requests

from . import get_data_dir


# Hack it up to find the filenames ending with .h5
class MyHTMLParser(HTMLParser):
    """HTML parser class that uses the HTMLParser to parse a starttag.

    See Also
    --------
    html.parser.HTMLParser

    Examples
    --------
    To instantiate a MyHTMLParser instance:

    parser = MyHTMLParser()
    parser.handle_starttag(tag, attrs)
    """

    def handle_starttag(self, tag, attrs):
        """
        Handle the start tag of an element (e.g. <div id="main">).

        Parameters
        ----------
        tag : `str`
            The name of the tag converted to lower case.
        attrs : `list`
            A list of (name, value) pairs containing the attributes
            found inside the tag’s <> brackets
        """
        try:
            self.filenames
        except AttributeError:
            setattr(self, "filenames", [])
        if tag == "a":
            if attrs[0][0] == "href":
                if attrs[0][1].endswith(".h5"):
                    self.filenames.append(attrs[0][1])


def rs_download_sky():
    """Download sky files."""
    parser = argparse.ArgumentParser(
        description="Download precomputed skybrightness files for rubin_sim package"
    )
    parser.add_argument(
        "-f",
        "--force",
        dest="force",
        default=False,
        action="store_true",
        help="Force re-download of sky brightness data.",
    )
    parser.add_argument(
        "--url_base",
        type=str,
        default="https://s3df.slac.stanford.edu/groups/rubin/static/sim-data/sims_skybrightness_pre/h5_2023_09_12/",
        help="Root URL of download location",
    )
    args = parser.parse_args()

    data_dir = get_data_dir()
    destination = os.path.join(data_dir, "skybrightness_pre")
    if not os.path.isdir(data_dir):
        os.mkdir(data_dir)
    if not os.path.isdir(destination):
        os.mkdir(destination)

    # Get the index file
    r = requests.get(args.url_base)
    # Find the filenames
    parser = MyHTMLParser()
    parser.feed(r.text)
    parser.close()
    # Copy the sky data files, if they're not already present
    for file in parser.filenames:
        if not os.path.isfile(os.path.join(destination, file)) or args.force:
            url = args.url_base + file
            print(f"Downloading file {file} from {url}")
            r = requests.get(url)
            with open(os.path.join(destination, file), "wb") as f:
                f.write(r.content)



__all__ = ("SeeingModel",)

import warnings

import numpy as np

from rubin_scheduler.utils import SysEngVals


class SeeingModel:
    """LSST FWHM calculations for FWHM_effective and FWHM_geometric.

    Calculations of the delivered values are based on equations in
    Document-20160 ("Atmospheric and Delivered Image Quality in OpSim"
    by Bo Xin, George Angeli, Zeljko Ivezic)
    An example of the calculation of delivered image seeing from DIMM
    FWHM_500 is available in
    https://smtn-002.lsst.io/
    #calculating-m5-values-in-the-lsst-operations-simulator

    Parameters
    ----------
    filter_list : `list` [`str`], opt
        List of the filter bandpasses for which to calculate delivered
        FWHM_effective and FWHM_geometric
        Default ['u', 'g', 'r', 'i', 'z', 'y']
    eff_wavelens : `list` [`float`] or None, opt
        Effective wavelengths for those bandpasses, in nanometers.
        If None, loades from rubin_scheduler.utils.SysEngVals
    telescope_seeing : `float`, opt
        The contribution to the delivered FWHM from the telescope,
        in arcseconds. Default 0.25"
    optical_design_seeing : `float`, opt
        The contribution to the seeing from the optical design, in
        arcseconds. Default 0.08 arcseconds
    camera_seeing : `float`, opt
        The contribution to the seeing from the camera, in arcseconds.
        Default 0.30 arcseconds
    raw_seeing_wavelength : `float`, opt
        The wavelength of the DIMM-delivered equivalent FWHM, in
        nanometers. Default 500nm.
    efd_seeing : `str`, opt
        The name of the DIMM FWHM measurements in the efd /
        conditions object. Default `FWHM_500`
    """

    def __init__(
        self,
        filter_list=["u", "g", "r", "i", "z", "y"],
        eff_wavelens=None,
        telescope_seeing=0.25,
        optical_design_seeing=0.08,
        camera_seeing=0.30,
        raw_seeing_wavelength=500,
        efd_seeing="FWHM_500",
    ):
        self.filter_list = filter_list
        if eff_wavelens is None:
            sev = SysEngVals()
            eff_wavelens = [sev.eff_wavelengths[f] for f in filter_list]
        self.eff_wavelens = np.array(eff_wavelens)
        self.telescope_seeing = telescope_seeing
        self.optical_design_seeing = optical_design_seeing
        self.raw_seeing_wavelength = raw_seeing_wavelength
        self.efd_seeing = efd_seeing
        self.camera_seeing = camera_seeing

        self._set_fwhm_zenith_system()

    def configure(self):
        """Deprecated. Configure through the init method."""
        warnings.warn("the configure method is deprecated")

    def config_info(self):
        """Deprecated. Report configuration parameters and version
        information."""
        warnings.warn("the config_info method is deprecated.")

    def _set_fwhm_zenith_system(self):
        """Calculate the system contribution to FWHM at zenith.

        This is simply the individual telescope, optics, and camera
        contributions combined in quadrature.
        """
        self.fwhm_system_zenith = np.sqrt(
            self.telescope_seeing**2 + self.optical_design_seeing**2 + self.camera_seeing**2
        )

    def __call__(self, fwhm_z, airmass):
        """Calculate the seeing values FWHM_eff and FWHM_geom at the
        given airmasses, for the specified effective wavelengths, given
        FWHM_zenith (typically FWHM_500).

        FWHM_geom represents the geometric size of the PSF; FWHM_eff
        represents the FWHM of a single gaussian which encloses the
        same number of pixels as N_eff (the number of pixels enclosed
        in the actual PSF -- this is the value to use when calculating
        SNR).

        FWHM_geom(") = 0.822 * FWHM_eff(") + 0.052"

        The FWHM_eff includes a contribution from the system and from
        the atmosphere.  Both of these are expected to scale with
        airmass^0.6 and with (500(nm)/wavelength(nm))^0.3.
        FWHM_eff = 1.16 * sqrt(FWHM_sys**2 + 1.04*FWHM_atm**2)

        Parameters
        ----------
        fwhm_z: `float`, or efdData `dict`
            FWHM at zenith (arcsec).
        airmass: `float`, `np.array`, or targetDict `dict`
            Airmass (unitless).

        Returns
        -------
        FWHMeff, FWHMGeom : `dict` of {`numpy.ndarray`, `numpy.ndarray`}
            FWHMeff, FWHMgeom: both are the same shape numpy.ndarray.
            If airmass is a single value, FWHMeff & FWHMgeom are 1-d
            arrays, with the same order as eff_wavelen (i.e.
            eff_wavelen[0] = u, then FWHMeff[0] = u). If airmass is a
            numpy array, FWHMeff and FWHMgeom are 2-d arrays, in the
            order of <filter><airmass> (i.e. eff_wavelen[0] = u, 1-d
            array over airmass range).
        """
        if isinstance(fwhm_z, dict):
            fwhm_z = fwhm_z[self.efd_seeing]
        if isinstance(airmass, dict):
            airmass = airmass["airmass"]
        airmass_correction = np.power(airmass, 0.6)
        wavelen_correction = np.power(self.raw_seeing_wavelength / self.eff_wavelens, 0.3)
        if isinstance(airmass, np.ndarray):
            fwhm_system = self.fwhm_system_zenith * np.outer(
                np.ones(len(wavelen_correction)), airmass_correction
            )
            fwhm_atmo = fwhm_z * np.outer(wavelen_correction, airmass_correction)
        else:
            fwhm_system = self.fwhm_system_zenith * airmass_correction
            fwhm_atmo = fwhm_z * wavelen_correction * airmass_correction
        # Calculate combined FWHMeff.
        fwhm_eff = 1.16 * np.sqrt(fwhm_system**2 + 1.04 * fwhm_atmo**2)
        # Translate to FWHMgeom.
        fwhm_geom = self.fwhm_eff_to_fwhm_geom(fwhm_eff)
        return {"fwhmEff": fwhm_eff, "fwhmGeom": fwhm_geom}

    @staticmethod
    def fwhm_eff_to_fwhm_geom(fwhm_eff):
        """Calculate FWHM_geom from FWHM_eff.

        Parameters
        ----------
        fwhm_eff : `float` or `np.ndarray`

        Returns
        -------
        FWHM_geom : `float` or `np.ndarray`
        """
        return 0.822 * fwhm_eff + 0.052

    @staticmethod
    def fwhm_geom_to_fwhm_eff(fwhm_geom):
        """Calculate FWHM_eff from FWHM_geom.

        Parameters
        ----------
        fwhm_geom : `float` or `np.ndarray`

        Returns
        -------
        FWHM_eff : `float` or `np.ndarray`
        """
        return (fwhm_geom - 0.052) / 0.822
